# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################


import requests
from keycloak.openid_connect import KeycloakOpenidConnect
from keycloak.realm import KeycloakRealm

import sickrage


class AuthServer(object):
    __server = {}
    __client = {}

    def __init__(self):
        self.server_url = 'https://auth.sickrage.ca'
        self.server_realm = 'sickrage'
        self.client_id = 'sickrage-app'
        self._certs = None

    @property
    def client(self):
        return self.__get_client()

    @property
    def health(self):
        for i in range(3):
            try:
                health = requests.get("{base}/auth/realms/{realm}".format(base=self.server_url, realm=self.server_realm), verify=False, timeout=30).ok
            except (requests.exceptions.ConnectionError, requests.exceptions.ReadTimeout):
                pass
            else:
                break
        else:
            health = False

        if not health:
            sickrage.app.log.debug("SiCKRAGE authorization server is currently unreachable")
            return False

        return True

    def get_url(self, *args, **kwargs):
        try:
            return self.client.get_url(*args, **kwargs)
        except requests.exceptions.ConnectionError as e:
            return

    def certs(self):
        try:
            if not self._certs and self.health:
                self._certs = self.client.certs()
            return self._certs
        except requests.exceptions.ConnectionError as e:
            return

    def logout(self, *args, **kwargs):
        if not self.health:
            return

        return self.client.logout(*args, **kwargs)

    def decode_token(self, *args, **kwargs):
        return self.client.decode_token(*args, **kwargs)

    def refresh_token(self, *args, **kwargs):
        if not self.health:
            return

        return self.client.refresh_token(*args, **kwargs)

    def authorization_code(self, *args, **kwargs):
        if not self.health:
            return

        return self.client.authorization_code(*args, **kwargs)

    def authorization_url(self, **kwargs):
        if not self.health:
            return

        return self.client.authorization_url(**kwargs)

    def token_exchange(self, access_token, scope='offline_access'):
        if not self.health:
            return

        exchange = {'scope': scope, 'subject_token': access_token}
        return self.client.token_exchange(**exchange)

    def __get_client(self) -> KeycloakOpenidConnect:
        client = self.__client.get('client', None)
        if client is None:
            self.__client.update({'client': self.__get_server().open_id_connect(self.client_id, '')})
            client = self.__client.get('client', None)
        return client

    def __get_server(self) -> KeycloakRealm:
        server = self.__server.get('server', None)
        if server is None:
            self.__server.update({'server': KeycloakRealm(server_url=self.server_url, realm_name=self.server_realm)})
            server = self.__server.get('server', None)
        return server



# Author: Marvin Pinto <me@marvinp.ca>
# Author: Dennis Lutter <lad1337@gmail.com>
# Author: Aaron Bieber <deftly@gmail.com>
# URL: https://sickrage.ca
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
from urllib import parse

import requests

import sickrage
from sickrage.core.websession import WebSession
from sickrage.notification_providers import NotificationProvider


class FreeMobileNotification(NotificationProvider):
    def __init__(self):
        super(FreeMobileNotification, self).__init__()
        self.name = 'freemobile'

    def test_notify(self, id=None, apiKey=None):
        return self._notifyFreeMobile('Test', "This is a test notification from SiCKRAGE", id, apiKey, force=True)

    def _sendFreeMobileSMS(self, title, msg, id=None, apiKey=None):
        """
        Sends a SMS notification

        msg: The message to send (unicode)
        title: The title of the message
        userKey: The pushover user id to send the message to (or to subscribe with)

        returns: True if the message succeeded, False otherwise
        """

        if id is None:
            id = sickrage.app.config.freemobile.user_id

        if apiKey is None:
            apiKey = sickrage.app.config.freemobile.apikey

        sickrage.app.log.debug("Free Mobile in use with API KEY: " + apiKey)

        # build up the URL and parameters
        msg = msg.strip()
        msg_quoted = parse.quote(title + ": " + msg)
        URL = "https://smsapi.free-mobile.fr/sendmsg?user=" + id + "&pass=" + apiKey + "&msg=" + msg_quoted

        # send the request to Free Mobile
        try:
            WebSession().get(URL)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 400:
                message = "Missing parameter(s)."
                sickrage.app.log.error(message)
                return False, message
            if e.response.status_code == 402:
                message = "Too much SMS sent in a short time."
                sickrage.app.log.error(message)
                return False, message
            if e.response.status_code == 403:
                message = "API service isn't enabled in your account or ID / API key is incorrect."
                sickrage.app.log.error(message)
                return False, message
            if e.response.status_code == 500:
                message = "Server error. Please retry in few moment."
                sickrage.app.log.error(message)
                return False, message

            message = "Error while sending SMS: {}".format(e)
            sickrage.app.log.error(message)
            return False, message

        message = "Free Mobile SMS successful."
        sickrage.app.log.info(message)
        return True, message

    def notify_snatch(self, ep_name, title=None):
        if not title:
            title = self.notifyStrings[self.NOTIFY_SNATCH]

        if sickrage.app.config.freemobile.notify_on_snatch:
            self._notifyFreeMobile(title, ep_name)

    def notify_download(self, ep_name, title=None):
        if not title:
            title = self.notifyStrings[self.NOTIFY_DOWNLOAD]

        if sickrage.app.config.freemobile.notify_on_download:
            self._notifyFreeMobile(title, ep_name)

    def notify_subtitle_download(self, ep_name, lang, title=None):
        if not title:
            title = self.notifyStrings[self.NOTIFY_SUBTITLE_DOWNLOAD]

        if sickrage.app.config.freemobile.notify_on_subtitle_download:
            self._notifyFreeMobile(title, ep_name + ": " + lang)

    def notify_version_update(self, new_version="??"):
        if sickrage.app.config.freemobile.enable:
            update_text = self.notifyStrings[self.NOTIFY_GIT_UPDATE_TEXT]
            title = self.notifyStrings[self.NOTIFY_GIT_UPDATE]
            self._notifyFreeMobile(title, update_text + new_version)

    def _notifyFreeMobile(self, title, message, id=None, apiKey=None, force=False):
        """
        Sends a SMS notification

        title: The title of the notification to send
        message: The message string to send
        id: Your Free Mobile customer ID
        apikey: Your Free Mobile API key
        force: Enforce sending, for instance for testing
        """

        if not sickrage.app.config.freemobile.enable and not force:
            sickrage.app.log.debug("Notification for Free Mobile not enabled, skipping this notification")
            return False, "Disabled"

        sickrage.app.log.debug("Sending a SMS for " + message)

        return self._sendFreeMobileSMS(title, message, id, apiKey)



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################
from marshmallow import fields
from marshmallow.validate import OneOf

from sickrage.core.enums import ProcessMethod
from sickrage.core.webserver.handlers.api.schemas import BaseSchema, BaseSuccessSchema


class PostProcessSchema(BaseSchema):
    """Complete postprocess schema"""

    path = fields.String(
        required=False,
        description="The path to the folder to post-process",
    )
    nzbName = fields.String(
        required=False,
        description="Release / NZB name if available",
    )
    processMethod = fields.String(
        required=False,
        default="copy",
        description="How should valid post-processed files be handled",
        example="copy",
        validate=[OneOf(choices=[x.name.lower() for x in ProcessMethod])]
    )
    type = fields.String(
        required=False,
        default="manual",
        description="The type of post-process being requested",
        example="auto",
        validate=[OneOf(choices=["auto", "manual"])]
    )
    delete = fields.Boolean(
        required=False,
        default=False,
        description="Mark download as failed",
    )
    failed = fields.Boolean(
        required=False,
        default=False,
        description="Mark download as failed",
    )
    isPriority = fields.Boolean(
        required=False,
        default=False,
        description="Replace the file even if it exists in a higher quality",
    )
    returnData = fields.Boolean(
        required=False,
        default=False,
        description="Returns the result of the post-process",
    )
    forceReplace = fields.Boolean(
        required=False,
        default=False,
        description="Force already post-processed files to be post-processed again",
    )
    forceNext = fields.Boolean(
        required=False,
        default=False,
        description="Waits for the current processing queue item to finish and returns result of this request",
    )


class PostProcessSuccessSchema(BaseSuccessSchema):
    data = fields.String(
        required=True,
        description="Validated and post-process logs",
    )


# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################


from tornado.web import authenticated

import sickrage
from sickrage.core.tv.show.history import History
from sickrage.core.webserver.handlers.base import BaseHandler


class HistoryHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        limit = self.get_argument('limit', None)

        if limit is None:
            if sickrage.app.config.gui.history_limit:
                limit = int(sickrage.app.config.gui.history_limit)
            else:
                limit = 100
        else:
            limit = int(limit)

        if sickrage.app.config.gui.history_limit != limit:
            sickrage.app.config.gui.history_limit = limit
            sickrage.app.config.save()

        compact = []

        for row in History().get(limit):
            action = {
                'action': row['action'],
                'provider': row['provider'],
                'release_group': row['release_group'],
                'resource': row['resource'],
                'time': row['date']
            }

            if not any((history['series_id'] == row['series_id'] and
                        history['season'] == row['season'] and
                        history['episode'] == row['episode'] and
                        history['quality'] == row['quality']) for history in compact):

                history = {
                    'actions': [action],
                    'quality': row['quality'],
                    'resource': row['resource'],
                    'season': row['season'],
                    'episode': row['episode'],
                    'series_id': row['series_id'],
                    'show_name': row['show_name']
                }

                compact.append(history)
            else:
                index = [i for i, item in enumerate(compact)
                         if item['series_id'] == row['series_id'] and
                         item['season'] == row['season'] and
                         item['episode'] == row['episode'] and
                         item['quality'] == row['quality']][0]

                history = compact[index]
                history['actions'].append(action)

                history['actions'].sort(key=lambda d: d['time'], reverse=True)

        submenu = [
            {'title': _('Clear History'), 'path': '/history/clear', 'icon': 'fas fa-trash',
             'class': 'clearhistory', 'confirm': True},
            {'title': _('Trim History'), 'path': '/history/trim', 'icon': 'fas fa-cut',
             'class': 'trimhistory', 'confirm': True},
        ]

        return self.render('history.mako',
                           historyResults=History().get(limit),
                           compactResults=compact,
                           limit=limit,
                           submenu=submenu,
                           title=_('History'),
                           header=_('History'),
                           topmenu="history",
                           controller='root',
                           action='history')


class HistoryClearHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        History().clear()
        sickrage.app.alerts.message(_('History cleared'))
        return self.redirect("/history/")


class HistoryTrimHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        History().trim()
        sickrage.app.alerts.message(_('Removed history entries older than 30 days'))
        return self.redirect("/history/")



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################


import sickrage
from sickrage.core.caches.tv_cache import TVCache
from sickrage.core.helpers import try_int
from sickrage.search_providers import TorrentProvider


class ZooqleProvider(TorrentProvider):
    def __init__(self):
        """Initialize the class."""
        super(ZooqleProvider, self).__init__('Zooqle', 'https://zooqle.com', False)

        # URLs
        self._urls.update({
            'search': '{base_url}/search'.format(**self._urls),
            'api': '{base_url}/api/media/%s'.format(**self._urls),
        })

        # custom settings
        self.custom_settings = {
            'minseed': 0,
            'minleech': 0
        }


        # Proper Strings
        self.proper_strings = ['PROPER', 'REPACK', 'REAL']

        # Cache
        self.cache = TVCache(self, min_time=15)

    def _get_torrent_info(self, torrent_hash):
        try:
            return self.session.get(self.urls['api'] % torrent_hash).json()
        except Exception:
            return {}

    def search(self, search_strings, age=0, series_id=None, series_provider_id=None, season=None, episode=None, **kwargs):
        """
        Search a provider and parse the results.

        :param search_strings: A dict with mode (key) and the search value (value)
        :param age: Not used
        :param ep_obj: Not used
        :returns: A list of search results (structure)
        """
        results = []

        # Search Params
        search_params = {
            'q': '* category:TV',
            's': 'dt',
            'v': 't',
            'sd': 'd',
        }

        for mode in search_strings:
            sickrage.app.log.debug('Search mode: {}'.format(mode))

            for search_string in search_strings[mode]:
                if mode != 'RSS':
                    sickrage.app.log.debug('Search string: {}'.format(search_string))
                    search_params['q'] = '{} category:TV'.format(search_string)

                search_params['fmt'] = 'rss'
                search_params['pg'] = 1

                while search_params['pg'] < 11:
                    data = self.cache.get_rss_feed(self.urls['search'], params=search_params)
                    if not data or not data.get('feed'):
                        sickrage.app.log.debug('No data returned from provider')
                        break

                    results += self.parse(data, mode)

                    total_results = try_int(data['feed'].get('opensearch_totalresults'))
                    start_index = try_int(data['feed'].get('opensearch_startindex'))
                    items_per_page = try_int(data['feed'].get('opensearch_itemsperpage'))
                    if not total_results or start_index + items_per_page > total_results:
                        break

                    search_params['pg'] += 1

        return results

    def parse(self, data, mode):
        """
        Parse search results for items.

        :param data: The raw response from a search
        :param mode: The current mode used to search, e.g. RSS

        :return: A list of items found
        """
        results = []

        if not data.get('entries'):
            sickrage.app.log.debug('Data returned from provider does not contain any torrents')
            return results

        for item in data['entries']:
            try:
                title = item.get('title')
                download_url = item.get('torrent_magneturi')
                if not all([title, download_url]):
                    continue

                seeders = try_int(item['torrent_seeds'])
                leechers = try_int(item['torrent_peers'])
                size = try_int(item['torrent_contentlength'], -1)

                results += [{
                    'title': title,
                    'link': download_url,
                    'size': size,
                    'seeders': seeders,
                    'leechers': leechers
                }]

                if mode != 'RSS':
                    sickrage.app.log.debug('Found result: {}'.format(title))
            except Exception:
                sickrage.app.log.error("Failed parsing provider")

        return results



# coding=utf-8
# URL: https://sickrage.ca
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE. If not, see <http://www.gnu.org/licenses/>.


from urllib.parse import urljoin

import sickrage
from sickrage.core.caches.tv_cache import TVCache
from sickrage.core.helpers import validate_url, try_int, bs4_parser, convert_size
from sickrage.search_providers import TorrentProvider


class SkyTorrents(TorrentProvider):
    def __init__(self):
        super(SkyTorrents, self).__init__('SkyTorrents', 'https://www.skytorrents.lol', False)
        self._urls.update({
            "rss": "{base_url}/top100".format(**self._urls)
        })

        # custom settings
        self.custom_settings = {
            'custom_url': '',
            'minseed': 0,
            'minleech': 0
        }

        self.cache = TVCache(self)

    def search(self, search_strings, age=0, series_id=None, series_provider_id=None, season=None, episode=None, **kwargs):
        results = []

        for mode in search_strings:
            sickrage.app.log.debug("Search Mode: {}".format(mode))
            for search_string in search_strings[mode]:
                if mode != "RSS":
                    sickrage.app.log.debug("Search string: {0}".format(search_string))

                search_url = (self.urls["base_url"], self.urls["rss"])[mode == "RSS"]

                if self.custom_settings['custom_url']:
                    if not validate_url(self.custom_settings['custom_url']):
                        sickrage.app.log.warning("Invalid custom url: {}".format(self.custom_settings['custom_url']))
                        return results
                    search_url = urljoin(self.custom_settings['custom_url'], search_url.split(self.urls['base_url'])[1])

                if mode != "RSS":
                    search_params = {'query': search_string, 'sort': ('seeders', 'created')[mode == 'RSS'], 'type': 'video', 'tag': 'hd', 'category': 'show'}
                else:
                    search_params = {'category': 'show', 'type': 'video', 'sort': 'created'}

                resp = self.session.get(search_url, params=search_params)
                if not resp or not resp.text:
                    sickrage.app.log.debug("No data returned from provider")
                    continue

                results += self.parse(resp.text, mode)

        return results

    def parse(self, data, mode, **kwargs):
        """
        Parse search results from data
        :param data: response data
        :param mode: search mode
        :return: search results
        """

        results = []

        with bs4_parser(data) as html:
            labels = [label.get_text(strip=True) for label in html('th')]
            for item in html('tr', attrs={'data-size': True}):
                try:
                    cells = item.findChildren('td')

                    title_block_links = cells[labels.index('Name')].find_all('a')
                    title = title_block_links[0].get_text(strip=True)
                    download_url = title_block_links[2]['href']

                    seeders = try_int(cells[labels.index(('Seeders', 'Seeds')[mode == "RSS"])].get_text(strip=True))
                    leechers = try_int(cells[labels.index(('Leechers', 'Peers')[mode == "RSS"])].get_text(strip=True))

                    size = try_int(item['data-size'])

                    results += [
                        {'title': title, 'link': download_url, 'size': size, 'seeders': seeders, 'leechers': leechers}
                    ]

                    if mode != 'RSS':
                        sickrage.app.log.debug("Found result: {}".format(title))
                except Exception:
                    sickrage.app.log.error("Failed parsing provider")

        return results



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca/
# Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.


import datetime
import sys
from collections import deque

from sickrage.core.common import dateTimeFormat


class UIError(object):
    """
    Represents an error to be displayed in the web UI.
    """

    def __init__(self, message):
        self.time = datetime.datetime.now().strftime(dateTimeFormat)
        self.title = sys.exc_info()[-2] or message
        self.message = message


class UIWarning(object):
    """
    Represents an error to be displayed in the web UI.
    """

    def __init__(self, message):
        self.time = datetime.datetime.now().strftime(dateTimeFormat)
        self.title = sys.exc_info()[-2] or message
        self.message = message


class ErrorViewer(object):
    """
    Keeps a static list of UIErrors to be displayed on the UI and allows
    the list to be cleared.
    """

    def __init__(self):
        self.errors = deque(maxlen=100)

    def add(self, error, ui=False):
        self.errors += [(error, UIError(error))[ui]]

    def clear(self):
        self.errors.clear()

    def get(self, *args, **kwargs):
        return self.errors

    def count(self):
        return len(self.errors)


class WarningViewer(object):
    """
    Keeps a static list of (warning) UIErrors to be displayed on the UI and allows
    the list to be cleared.
    """

    def __init__(self):
        self.warnings = deque(maxlen=100)

    def add(self, warning, ui=False):
        self.warnings += [(warning, UIWarning(warning))[ui]]

    def clear(self):
        self.warnings.clear()

    def get(self, *args, **kwargs):
        return self.warnings

    def count(self):
        return len(self.warnings)



import os
import shutil

import glob
from setuptools import setup, Command

from sickrage import __version__


def requirements():
    with open(os.path.abspath(os.path.join(os.path.dirname(__file__), 'requirements.txt'))) as f:
        return f.read().splitlines()


def requirements_dev():
    with open(os.path.abspath(os.path.join(os.path.dirname(__file__), 'requirements-dev.txt'))) as f:
        return f.read().splitlines()


class CleanCommand(Command):
    """Custom clean command to tidy up the project root."""
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        shutil.rmtree(os.path.abspath(os.path.join(os.path.dirname(__file__), 'build')), ignore_errors=True)
        shutil.rmtree(os.path.abspath(os.path.join(os.path.dirname(__file__), '*.pyc')), ignore_errors=True)
        shutil.rmtree(os.path.abspath(os.path.join(os.path.dirname(__file__), '*.tgz')), ignore_errors=True)
        shutil.rmtree(os.path.abspath(os.path.join(os.path.dirname(__file__), 'sickrage.egg-info')), ignore_errors=True)
        [os.remove(f) for f in glob.glob("dist/sickrage-*")]


cmd_class = {'clean': CleanCommand}

# Check for Babel availability
try:
    from babel.messages.frontend import compile_catalog, extract_messages, init_catalog, update_catalog

    cmd_class.update(dict(
        compile_catalog=compile_catalog,
        extract_messages=extract_messages,
        init_catalog=init_catalog,
        update_catalog=update_catalog
    ))
except ImportError:
    pass

setup(
    name='sickrage',
    version=__version__,
    description='Automatic Video Library Manager for TV Shows',
    author='echel0n',
    author_email='echel0n@sickrage.ca',
    license='GPLv3',
    url='https://git.sickrage.ca',
    keywords=['sickrage', 'sickragetv', 'tv', 'torrent', 'nzb', 'video', 'echel0n'],
    packages=['sickrage'],
    install_requires=requirements(),
    extras_require={
        'dev': requirements_dev()
    },
    include_package_data=True,
    python_requires='>=3',
    platforms='any',
    zip_safe=False,
    test_suite='tests',
    cmdclass=cmd_class,
    entry_points={
        "console_scripts": [
            "sickrage=sickrage:main"
        ]
    },
    message_extractors={
        'sickrage/core/webserver': [
            ('**/views/**.mako', 'mako', {'input_encoding': 'utf-8'})
        ],
        'sickrage': [
            ('**.py', 'python', None)
        ],
        'src': [
            ('**/js/*.min.js', 'ignore', None),
            ('**/js/*.js', 'javascript', {'input_encoding': 'utf-8'})
        ],
    }
)



from __future__ import absolute_import, division, print_function

import calendar
import datetime
import logging
import socket
import time
from threading import RLock

import requests
from requests.adapters import DEFAULT_POOLBLOCK, HTTPAdapter

from trakt.core.configuration import DEFAULT_HTTP_RETRY, DEFAULT_HTTP_MAX_RETRIES, DEFAULT_HTTP_TIMEOUT, \
    DEFAULT_HTTP_RETRY_SLEEP
from trakt.core.context_stack import ContextStack
from trakt.core.helpers import synchronized
from trakt.core.keylock import KeyLock
from trakt.core.request import TraktRequest

try:
    import ssl
except ImportError:
    ssl = None

log = logging.getLogger(__name__)


class HttpClient(object):
    def __init__(self, client, adapter_kwargs=None, keep_alive=True):
        self.client = client

        self.adapter_kwargs = adapter_kwargs or {}
        self.keep_alive = keep_alive

        # Build client
        self.configuration = ContextStack()
        self.session = None

        self._proxies = {}
        self._ssl_version = None

        self._oauth_refreshing = KeyLock()
        self._oauth_validate_lock = RLock()

        # Build requests session
        self.rebuild()

    @property
    def proxies(self):
        if self.session and self.session.proxies:
            return self.session.proxies

        return self._proxies

    @proxies.setter
    def proxies(self, proxies):
        if self.session:
            self.session.proxies = proxies

        self._proxies = proxies

    @property
    def ssl_version(self):
        return self._ssl_version

    @ssl_version.setter
    def ssl_version(self, version):
        self._ssl_version = version

        # Rebuild session (to apply ssl version change)
        self.rebuild()

    def configure(self, path=None):
        self.configuration.push(base_path=path)

        return self

    def request(self, method, path=None, params=None, data=None, query=None, authenticated=False,
                validate_token=True, **kwargs):

        # Retrieve configuration
        ctx = self.configuration.pop()

        # Build request
        request = TraktRequest(
            self.client,
            method=method,

            path=self._build_path(ctx, path),
            params=params,

            data=data,
            query=query,

            authenticated=authenticated,
            **kwargs
        )

        # Validate authentication details (OAuth)
        if authenticated and validate_token and not self.validate():
            return None

        # Prepare request
        prepared = request.prepare()

        if not self.keep_alive:
            prepared.headers['Connection'] = 'close'

        # Send request
        return self.send(prepared)

    def send(self, request):
        # Retrieve http configuration
        retry = self.client.configuration.get('http.retry', DEFAULT_HTTP_RETRY)
        max_retries = self.client.configuration.get('http.max_retries', DEFAULT_HTTP_MAX_RETRIES)
        retry_sleep = self.client.configuration.get('http.retry_sleep', DEFAULT_HTTP_RETRY_SLEEP)
        timeout = self.client.configuration.get('http.timeout', DEFAULT_HTTP_TIMEOUT)

        # Send request
        response = None

        for i in range(max_retries + 1):
            if i > 0:
                log.warn('Retry # %s', i)

            # Send request
            try:
                response = self.session.send(request, timeout=timeout)
            except socket.gaierror as e:
                code, __ = e

                if code != 8:
                    raise e

                log.warn('Encountered socket.gaierror (code: 8)')

                response = self.rebuild().send(request, timeout=timeout)

            # Retry requests on errors >= 500 (when enabled)
            if not retry or response.status_code < 500:
                break

            log.warn('Continue retry since status is %s, waiting %s seconds', response.status_code, retry_sleep)
            time.sleep(retry_sleep)

        return response

    def delete(self, path=None, params=None, data=None, **kwargs):
        return self.request('DELETE', path, params, data, **kwargs)

    def get(self, path=None, params=None, data=None, **kwargs):
        return self.request('GET', path, params, data, **kwargs)

    def post(self, path=None, params=None, data=None, **kwargs):
        return self.request('POST', path, params, data, **kwargs)

    def put(self, path=None, params=None, data=None, **kwargs):
        return self.request('PUT', path, params, data, **kwargs)

    def rebuild(self):
        if self.session:
            log.info('Rebuilding session and connection pools...')

        # Build the connection pool
        self.session = requests.Session()
        self.session.proxies = self.proxies

        # Mount adapters
        self.session.mount('http://', HTTPAdapter(**self.adapter_kwargs))
        self.session.mount('https://', HTTPSAdapter(ssl_version=self._ssl_version, **self.adapter_kwargs))

        return self.session

    def validate(self):
        config = self.client.configuration

        # xAuth
        if config['auth.login'] and config['auth.token']:
            return True

        # OAuth
        if config['oauth.token']:
            # Validate OAuth token, refresh if needed
            return self._validate_oauth()

        return False

    def _build_path(self, ctx, path):
        if not ctx:
            # No context available
            return path

        if ctx.base_path and path:
            # Prepend `base_path` to relative `path`s
            if not path.startswith('/'):
                path = ctx.base_path + '/' + path
        elif ctx.base_path:
            # Set path to `base_path
            path = ctx.base_path

        return path

    @synchronized(lambda self: self._oauth_validate_lock)
    def _validate_oauth(self):
        config = self.client.configuration

        # Ensure token expiry is available
        if config['oauth.created_at'] is None or config['oauth.expires_in'] is None:
            log.debug('OAuth - Missing "created_at" or "expires_in" parameters, '
                      'unable to determine if the current token is still valid')
            return True

        # Calculate expiry
        current = calendar.timegm(datetime.datetime.utcnow().utctimetuple())
        expires_at = config['oauth.created_at'] + config['oauth.expires_in'] - (48 * 60 * 60)

        if current < expires_at:
            return True

        if not config['oauth.refresh']:
            log.warn('OAuth - Unable to refresh expired token (token refreshing hasn\'t been enabled)')
            return False

        if not config['oauth.refresh_token']:
            log.warn('OAuth - Unable to refresh expired token ("refresh_token" parameter hasn\'t been defined)')
            return False

        # Retrieve username
        username = config['oauth.username']

        if not username:
            log.info('OAuth - Current username is not available ("username" parameter hasn\'t been defined)')

        # Acquire refreshing lock
        if not self._oauth_refreshing[username].acquire(False):
            log.warn('OAuth - Token is already being refreshed for %r', username)
            return False

        log.info('OAuth - Token has expired, refreshing token...')

        # Refresh token
        try:
            if not self._refresh_oauth():
                return False

            log.info('OAuth - Token has been refreshed')
            return True
        finally:
            # Release refreshing lock
            self._oauth_refreshing[username].release()

    def _refresh_oauth(self):
        config = self.client.configuration

        # Refresh token
        response = self.client['oauth'].token_refresh(
            config['oauth.refresh_token'], 'urn:ietf:wg:oauth:2.0:oob',
            parse=False
        )

        if response is None:
            log.warn('OAuth - Unable to refresh expired token (no response returned)')
            return False

        if response.status_code < 200 or response.status_code >= 300:
            # Clear current configuration
            config.current.oauth.clear()

            # Handle refresh rejection
            if response.status_code == 401:
                log.warn('OAuth - Unable to refresh expired token (rejected)')

                # Fire rejected event
                self.client.emit('oauth.refresh.rejected', config['oauth.username'])
                return False

            # Unknown error returned
            log.warn('OAuth - Unable to refresh expired token (code: %r)', response.status_code)
            return False

        # Retrieve authorization parameters from response
        authorization = response.json()

        # Update current configuration
        config.current.oauth.from_response(
            authorization,
            refresh=config['oauth.refresh'],
            username=config['oauth.username']
        )

        # Fire refresh event
        self.client.emit('oauth.refresh', config['oauth.username'], authorization)

        # Fire legacy refresh event
        self.client.emit('oauth.token_refreshed', authorization)
        return True


class HTTPSAdapter(HTTPAdapter):
    def __init__(self, ssl_version=None, *args, **kwargs):
        self._ssl_version = ssl_version

        super(HTTPSAdapter, self).__init__(*args, **kwargs)

    def init_poolmanager(self, connections, maxsize, block=DEFAULT_POOLBLOCK, **pool_kwargs):
        pool_kwargs['ssl_version'] = self._ssl_version

        return super(HTTPSAdapter, self).init_poolmanager(
            connections, maxsize, block,
            **pool_kwargs
        )



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
import datetime

from sqlalchemy import Column, Integer, Text, ForeignKeyConstraint, String, DateTime, Boolean, Index, Date, BigInteger, func, literal_column, Enum, exists
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

import sickrage
from sickrage.core.common import Qualities, EpisodeStatus
from sickrage.core.databases import SRDatabase, SRDatabaseBase, IntFlag
from sickrage.core.enums import SearchFormat, SeriesProviderID


class MainDB(SRDatabase):
    base = declarative_base(cls=SRDatabaseBase)

    def __init__(self, db_type, db_prefix, db_host, db_port, db_username, db_password):
        super(MainDB, self).__init__('main', db_type, db_prefix, db_host, db_port, db_username, db_password)

    def initialize(self):
        self.base.metadata.create_all(self.engine)

    def cleanup(self):
        def remove_orphaned():
            session = self.session()

            # orphaned episode entries
            for orphaned_result in session.query(self.TVEpisode).filter(~exists().where(self.TVEpisode.series_id == self.TVShow.series_id)):
                sickrage.app.log.debug(f"Orphaned episode detected! episode_id: {orphaned_result.episode_id}")
                sickrage.app.log.info(f"Deleting orphaned episode with episode_id: {orphaned_result.episode_id}")
                session.delete(orphaned_result)
                session.commit()

            # orphaned imdb info entries
            for orphaned_result in session.query(self.IMDbInfo).filter(~exists().where(self.IMDbInfo.series_id == self.TVShow.series_id)):
                sickrage.app.log.debug(f"Orphaned imdb info detected! series_id: {orphaned_result.series_id}")
                sickrage.app.log.info(f"Deleting orphaned imdb info with series_id: {orphaned_result.series_id}")
                session.delete(orphaned_result)
                session.commit()

            # orphaned series provider mapping entries
            for orphaned_result in session.query(self.SeriesProviderMapping).filter(~exists().where(self.SeriesProviderMapping.series_id == self.TVShow.series_id)):
                sickrage.app.log.debug(f"Orphaned series provider mapper detected! series_id: {orphaned_result.series_id}")
                sickrage.app.log.info(f"Deleting orphaned series provider mapper with series_id: {orphaned_result.series_id}")
                session.delete(orphaned_result)
                session.commit()

            # orphaned whitelist entries
            for orphaned_result in session.query(self.Whitelist).filter(~exists().where(self.Whitelist.series_id == self.TVShow.series_id)):
                sickrage.app.log.debug(f"Orphaned whitelist detected! series_id: {orphaned_result.series_id}")
                sickrage.app.log.info(f"Deleting orphaned whitelist with series_id: {orphaned_result.series_id}")
                session.delete(orphaned_result)
                session.commit()

            # orphaned blacklist entries
            for orphaned_result in session.query(self.Blacklist).filter(~exists().where(self.Blacklist.series_id == self.TVShow.series_id)):
                sickrage.app.log.debug(f"Orphaned blacklist detected! series_id: {orphaned_result.series_id}")
                sickrage.app.log.info(f"Deleting orphaned blacklist with series_id: {orphaned_result.series_id}")
                session.delete(orphaned_result)
                session.commit()

            # orphaned history entries
            for orphaned_result in session.query(self.History).filter(~exists().where(self.History.series_id == self.TVShow.series_id)):
                sickrage.app.log.debug(f"Orphaned history detected! series_id: {orphaned_result.series_id}")
                sickrage.app.log.info(f"Deleting orphaned history with series_id: {orphaned_result.series_id}")
                session.delete(orphaned_result)
                session.commit()

            # orphaned failed snatch history entries
            for orphaned_result in session.query(self.FailedSnatchHistory).filter(~exists().where(self.FailedSnatchHistory.series_id == self.TVShow.series_id)):
                sickrage.app.log.debug(f"Orphaned failed snatch history detected! series_id: {orphaned_result.series_id}")
                sickrage.app.log.info(f"Deleting orphaned failed snatch history with series_id: {orphaned_result.series_id}")
                session.delete(orphaned_result)
                session.commit()

            # orphaned failed snatch entries
            for orphaned_result in session.query(self.FailedSnatch).filter(~exists().where(self.FailedSnatch.series_id == self.TVShow.series_id)):
                sickrage.app.log.debug(f"Orphaned failed snatch detected! series_id: {orphaned_result.series_id}")
                sickrage.app.log.info(f"Deleting orphaned failed snatch with series_id: {orphaned_result.series_id}")
                session.delete(orphaned_result)
                session.commit()

        def remove_duplicate_shows():
            session = self.session()

            # count by series id
            duplicates = session.query(
                self.TVShow.series_id,
                func.count(self.TVShow.series_id).label('count')
            ).group_by(
                self.TVShow.series_id
            ).having(literal_column('count') > 1).all()

            for cur_duplicate in duplicates:
                sickrage.app.log.debug("Duplicate show detected! series_id: {dupe_id} count: {dupe_count}".format(dupe_id=cur_duplicate.series_id,
                                                                                                                  dupe_count=cur_duplicate.count))

                for result in session.query(self.TVShow).filter_by(series_id=cur_duplicate.series_id).limit(cur_duplicate.count - 1):
                    session.query(self.TVShow).filter_by(series_id=result.series_id).delete()
                    session.commit()

        def remove_duplicate_episodes():
            session = self.session()

            # count by season/episode
            duplicates = session.query(
                self.TVEpisode.series_id,
                self.TVEpisode.season,
                self.TVEpisode.episode,
                func.count(self.TVEpisode.episode_id).label('count')
            ).group_by(
                self.TVEpisode.series_id,
                self.TVEpisode.season,
                self.TVEpisode.episode,
            ).having(literal_column('count') > 1).all()

            for cur_duplicate in duplicates:
                sickrage.app.log.debug("Duplicate episode detected! "
                                       "series_id: {dupe_id} "
                                       "season: {dupe_season} "
                                       "episode {dupe_episode} count: {dupe_count}".format(dupe_id=cur_duplicate.series_id,
                                                                                           dupe_season=cur_duplicate.season,
                                                                                           dupe_episode=cur_duplicate.episode,
                                                                                           dupe_count=cur_duplicate.count))

                for result in session.query(self.TVEpisode).filter_by(series_id=cur_duplicate.series_id,
                                                                      season=cur_duplicate.season,
                                                                      episode=cur_duplicate.episode).limit(cur_duplicate.count - 1):
                    session.query(self.TVEpisode).filter_by(series_id=result.series_id, season=result.season, episode=result.episode).delete()
                    session.commit()

            # count by series id
            duplicates = session.query(
                self.TVEpisode.series_id,
                self.TVEpisode.episode_id,
                self.TVEpisode.season,
                self.TVEpisode.episode,
                func.count(self.TVEpisode.episode_id).label('count')
            ).group_by(
                self.TVEpisode.series_id,
                self.TVEpisode.episode_id
            ).having(literal_column('count') > 1).all()

            for cur_duplicate in duplicates:
                sickrage.app.log.debug("Duplicate episode detected! "
                                       "series_id: {dupe_id} "
                                       "season: {dupe_season} "
                                       "episode {dupe_episode} count: {dupe_count}".format(dupe_id=cur_duplicate.series_id,
                                                                                           dupe_season=cur_duplicate.season,
                                                                                           dupe_episode=cur_duplicate.episode,
                                                                                           dupe_count=cur_duplicate.count))

                for result in session.query(self.TVEpisode).filter_by(series_id=cur_duplicate.series_id,
                                                                      episode_id=cur_duplicate.episode_id).limit(cur_duplicate.count - 1):
                    session.query(self.TVEpisode).filter_by(series_id=result.series_id, episode_id=result.episode_id).delete()
                    session.commit()

        def fix_duplicate_episode_scene_numbering():
            session = self.session()

            duplicates = session.query(
                self.TVEpisode.series_id,
                self.TVEpisode.scene_season,
                self.TVEpisode.scene_episode,
                func.count(self.TVEpisode.series_id).label('count')
            ).group_by(
                self.TVEpisode.series_id,
                self.TVEpisode.scene_season,
                self.TVEpisode.scene_episode
            ).filter(
                self.TVEpisode.scene_season != -1,
                self.TVEpisode.scene_episode != -1
            ).having(literal_column('count') > 1)

            for cur_duplicate in duplicates:
                sickrage.app.log.debug("Duplicate episode scene numbering detected! "
                                       "series_id: {dupe_id} "
                                       "scene season: {dupe_scene_season} "
                                       "scene episode {dupe_scene_episode} count: {dupe_count}".format(dupe_id=cur_duplicate.series_id,
                                                                                                       dupe_scene_season=cur_duplicate.scene_season,
                                                                                                       dupe_scene_episode=cur_duplicate.scene_episode,
                                                                                                       dupe_count=cur_duplicate.count))

                for result in session.query(self.TVEpisode).filter_by(series_id=cur_duplicate.series_id,
                                                                      scene_season=cur_duplicate.scene_season,
                                                                      scene_episode=cur_duplicate.scene_episode).limit(cur_duplicate.count - 1):
                    result.scene_season = -1
                    result.scene_episode = -1
                    session.commit()

        def fix_duplicate_episode_scene_absolute_numbering():
            session = self.session()

            duplicates = session.query(
                self.TVEpisode.series_id,
                self.TVEpisode.scene_absolute_number,
                func.count(self.TVEpisode.series_id).label('count')
            ).group_by(
                self.TVEpisode.series_id,
                self.TVEpisode.scene_absolute_number
            ).filter(
                self.TVEpisode.scene_absolute_number != -1
            ).having(literal_column('count') > 1)

            for cur_duplicate in duplicates:
                sickrage.app.log.debug("Duplicate episode scene absolute numbering detected! "
                                       "series_id: {dupe_id} "
                                       "scene absolute number: {dupe_scene_absolute_number} "
                                       "count: {dupe_count}".format(dupe_id=cur_duplicate.series_id,
                                                                    dupe_scene_absolute_number=cur_duplicate.scene_absolute_number,
                                                                    dupe_count=cur_duplicate.count))

                for result in session.query(self.TVEpisode).filter_by(series_id=cur_duplicate.series_id,
                                                                      scene_absolute_number=cur_duplicate.scene_absolute_number). \
                        limit(cur_duplicate.count - 1):
                    result.scene_absolute_number = -1
                    session.commit()

        def remove_invalid_episodes():
            session = self.session()

            session.query(self.TVEpisode).filter_by(episode_id=0).delete()
            session.commit()

        def fix_invalid_scene_numbering():
            session = self.session()

            session.query(self.TVEpisode).filter_by(scene_season=0, scene_episode=0).update({
                self.TVEpisode.scene_season: -1,
                self.TVEpisode.scene_episode: -1
            })

            session.commit()

            session.query(self.TVEpisode).filter_by(scene_absolute_number=0).update({
                self.TVEpisode.scene_absolute_number: -1
            })

            session.commit()

            session.query(self.TVEpisode).filter(self.TVEpisode.season == self.TVEpisode.scene_season,
                                                 self.TVEpisode.episode == self.TVEpisode.scene_episode).update({
                self.TVEpisode.scene_season: -1,
                self.TVEpisode.scene_episode: -1
            })

            session.commit()

            session.query(self.TVEpisode).filter(self.TVEpisode.absolute_number == self.TVEpisode.scene_absolute_number).update({
                self.TVEpisode.scene_absolute_number: -1
            })

            session.commit()

        def fix_tvshow_table_columns():
            session = self.session()

            session.query(self.TVShow).filter_by(sub_use_sr_metadata=None).update({'sub_use_sr_metadata': False})
            session.query(self.TVShow).filter_by(skip_downloaded=None).update({'skip_downloaded': False})
            session.query(self.TVShow).filter_by(dvd_order=None).update({'dvd_order': False})
            session.query(self.TVShow).filter_by(subtitles=None).update({'subtitles': False})
            session.query(self.TVShow).filter_by(anime=None).update({'anime': False})
            session.query(self.TVShow).filter_by(flatten_folders=None).update({'flatten_folders': False})
            session.query(self.TVShow).filter_by(paused=None).update({'paused': False})
            session.query(self.TVShow).filter_by(last_xem_refresh=None).update({'last_xem_refresh': datetime.datetime.now()})

            session.commit()

        remove_orphaned()
        remove_duplicate_shows()
        remove_duplicate_episodes()
        remove_invalid_episodes()
        fix_invalid_scene_numbering()
        fix_duplicate_episode_scene_numbering()
        fix_duplicate_episode_scene_absolute_numbering()
        fix_tvshow_table_columns()

    class TVShow(base):
        __tablename__ = 'tv_shows'

        series_id = Column(Integer, index=True, primary_key=True)
        series_provider_id = Column(Enum(SeriesProviderID), index=True, primary_key=True)
        name = Column(Text, default='')
        location = Column(Text, default='')
        network = Column(Text, default='')
        genre = Column(Text, default='')
        overview = Column(Text, default='')
        classification = Column(Text, default='Scripted')
        runtime = Column(Integer, default=0)
        quality = Column(IntFlag(Qualities), default=Qualities.SD)
        airs = Column(Text, default='')
        status = Column(Text, default='')
        flatten_folders = Column(Boolean, nullable=False, default=0)
        paused = Column(Boolean, nullable=False, default=0)
        search_format = Column(Enum(SearchFormat), default=SearchFormat.STANDARD)
        scene = Column(Boolean, nullable=False, default=0)
        anime = Column(Boolean, nullable=False, default=0)
        subtitles = Column(Boolean, nullable=False, default=0)
        dvd_order = Column(Boolean, nullable=False, default=0)
        skip_downloaded = Column(Boolean, nullable=False, default=0)
        startyear = Column(Integer, default=0)
        lang = Column(Text, default='')
        imdb_id = Column(Text, default='')
        rls_ignore_words = Column(Text, default='')
        rls_require_words = Column(Text, default='')
        default_ep_status = Column(Enum(EpisodeStatus), default=EpisodeStatus.SKIPPED)
        sub_use_sr_metadata = Column(Boolean, nullable=False, default=0)
        notify_list = Column(Text, default='')
        search_delay = Column(Integer, default=0)
        scene_exceptions = Column(Text, default='')
        last_refresh = Column(DateTime(timezone=True), default=datetime.datetime.now())
        last_xem_refresh = Column(DateTime(timezone=True), default=datetime.datetime.now())
        last_scene_exceptions_refresh = Column(DateTime(timezone=True), default=datetime.datetime.now())
        last_update = Column(DateTime(timezone=True), default=datetime.datetime.now())
        last_backlog_search = Column(DateTime(timezone=True), default=datetime.datetime.now())
        last_proper_search = Column(DateTime(timezone=True), default=datetime.datetime.now())

        episodes = relationship('TVEpisode', uselist=True, backref='tv_shows', cascade="all, delete-orphan", lazy='dynamic')
        imdb_info = relationship('IMDbInfo', uselist=False, backref='tv_shows', cascade="all, delete-orphan")
        series_provider_mapping = relationship('SeriesProviderMapping', uselist=False, backref='tv_shows', cascade="all, delete-orphan")
        blacklist = relationship('Blacklist', uselist=False, backref='tv_shows', cascade="all, delete-orphan")
        whitelist = relationship('Whitelist', uselist=False, backref='tv_shows', cascade="all, delete-orphan")
        history = relationship('History', uselist=False, backref='tv_shows', cascade="all, delete-orphan")
        failed_snatch_history = relationship('FailedSnatchHistory', uselist=False, backref='tv_shows', cascade="all, delete-orphan")
        failed_snatches = relationship('FailedSnatch', uselist=False, backref='tv_shows', cascade="all, delete-orphan")

    class TVEpisode(base):
        __tablename__ = 'tv_episodes'
        __table_args__ = (
            ForeignKeyConstraint(['series_id', 'series_provider_id'], ['tv_shows.series_id', 'tv_shows.series_provider_id'], ondelete='CASCADE',
                                 name=f'fk_{__tablename__}_series_id_series_provider_id'),
            Index('idx_series_id_series_provider_id', 'series_id', 'series_provider_id'),
            Index('idx_series_id_episode_id', 'series_id', 'episode_id'),
            Index('idx_status_episode_airdate', 'status', 'episode', 'airdate'),
            Index('idx_season_episode_status_airdate', 'season', 'episode', 'status', 'airdate'),
            Index('idx_episode_id_airdate', 'episode_id', 'airdate'),
        )

        series_id = Column(Integer, index=True, primary_key=True)
        series_provider_id = Column(Enum(SeriesProviderID), index=True, primary_key=True)
        episode_id = Column(Integer, default=0)
        season = Column(Integer, index=True, primary_key=True)
        episode = Column(Integer, index=True, primary_key=True)
        absolute_number = Column(Integer, default=-1)
        scene_season = Column(Integer, default=-1)
        scene_episode = Column(Integer, default=-1)
        scene_absolute_number = Column(Integer, default=-1)
        xem_season = Column(Integer, default=-1)
        xem_episode = Column(Integer, default=-1)
        xem_absolute_number = Column(Integer, default=-1)
        name = Column(Text, default='')
        description = Column(Text, default='')
        subtitles = Column(Text, default='')
        subtitles_searchcount = Column(Integer, default=0)
        subtitles_lastsearch = Column(DateTime(timezone=True), default=func.current_timestamp())
        airdate = Column(Date, default=datetime.datetime.min)
        hasnfo = Column(Boolean, nullable=False, default=False)
        hastbn = Column(Boolean, nullable=False, default=False)
        status = Column(Enum(EpisodeStatus), default=EpisodeStatus.UNKNOWN)
        location = Column(Text, default='')
        file_size = Column(BigInteger, default=0)
        release_name = Column(Text, default='')
        is_proper = Column(Boolean, nullable=False, default=False)
        version = Column(Integer, default=-1)
        release_group = Column(Text, default='')

        show = relationship('TVShow', uselist=False, backref='tv_episodes')

    class IMDbInfo(base):
        __tablename__ = 'imdb_info'
        __table_args__ = (
            ForeignKeyConstraint(['series_id', 'imdb_id'], ['tv_shows.series_id', 'tv_shows.imdb_id'], ondelete='CASCADE',
                                 name=f'fk_{__tablename__}_series_id_imdb_id'),
        )

        series_id = Column(Integer, primary_key=True)
        imdb_id = Column(String(10), index=True, unique=True)
        rated = Column(Text)
        title = Column(Text)
        production = Column(Text)
        website = Column(Text)
        writer = Column(Text)
        actors = Column(Text)
        type = Column(Text)
        votes = Column(Text, nullable=False)
        seasons = Column(Text)
        poster = Column(Text)
        director = Column(Text)
        released = Column(Text)
        awards = Column(Text)
        genre = Column(Text, nullable=False)
        rating = Column(Text, nullable=False)
        language = Column(Text)
        country = Column(Text)
        runtime = Column(Text)
        metascore = Column(Text)
        year = Column(Text)
        plot = Column(Text)
        last_update = Column(DateTime(timezone=True), default=datetime.datetime.now())

    class SeriesProviderMapping(base):
        __tablename__ = 'series_provider_mapping'
        __table_args__ = (
            ForeignKeyConstraint(['series_id', 'series_provider_id'], ['tv_shows.series_id', 'tv_shows.series_provider_id'], ondelete='CASCADE',
                                 name=f'fk_{__tablename__}_series_id_series_provider_id'),
        )

        series_id = Column(Integer, primary_key=True)
        series_provider_id = Column(Enum(SeriesProviderID), primary_key=True)
        mapped_series_id = Column(Integer, nullable=False)
        mapped_series_provider_id = Column(Enum(SeriesProviderID), primary_key=True)

    class Blacklist(base):
        __tablename__ = 'blacklist'
        __table_args__ = (
            ForeignKeyConstraint(['series_id', 'series_provider_id'], ['tv_shows.series_id', 'tv_shows.series_provider_id'], ondelete='CASCADE',
                                 name=f'fk_{__tablename__}_series_id_series_provider_id'),
        )

        id = Column(Integer, autoincrement=True, primary_key=True)
        series_id = Column(Integer, nullable=False)
        series_provider_id = Column(Enum(SeriesProviderID), nullable=False)
        keyword = Column(Text, nullable=False)

    class Whitelist(base):
        __tablename__ = 'whitelist'
        __table_args__ = (
            ForeignKeyConstraint(['series_id', 'series_provider_id'], ['tv_shows.series_id', 'tv_shows.series_provider_id'], ondelete='CASCADE',
                                 name=f'fk_{__tablename__}_series_id_series_provider_id'),
        )

        id = Column(Integer, autoincrement=True, primary_key=True)
        series_id = Column(Integer, nullable=False)
        series_provider_id = Column(Enum(SeriesProviderID), nullable=False)
        keyword = Column(Text, nullable=False)

    class History(base):
        __tablename__ = 'history'
        __table_args__ = (
            ForeignKeyConstraint(['series_id', 'series_provider_id'], ['tv_shows.series_id', 'tv_shows.series_provider_id'], ondelete='CASCADE',
                                 name=f'fk_{__tablename__}_series_id_series_provider_id'),
        )

        id = Column(Integer, autoincrement=True, primary_key=True)
        series_id = Column(Integer, nullable=False)
        series_provider_id = Column(Enum(SeriesProviderID), nullable=False)
        season = Column(Integer, nullable=False)
        episode = Column(Integer, nullable=False)
        resource = Column(Text, nullable=False, index=True)
        action = Column(Integer, nullable=False)
        version = Column(Integer, default=-1)
        provider = Column(Text, nullable=False)
        date = Column(DateTime, nullable=False)
        quality = Column(IntFlag(Qualities), nullable=False)
        release_group = Column(Text, nullable=False)

    class FailedSnatchHistory(base):
        __tablename__ = 'failed_snatch_history'
        __table_args__ = (
            ForeignKeyConstraint(['series_id', 'series_provider_id'], ['tv_shows.series_id', 'tv_shows.series_provider_id'], ondelete='CASCADE',
                                 name=f'fk_{__tablename__}_series_id_series_provider_id'),
        )

        id = Column(Integer, autoincrement=True, primary_key=True)
        series_id = Column(Integer, nullable=False)
        series_provider_id = Column(Enum(SeriesProviderID), nullable=False)
        date = Column(DateTime, nullable=False)
        size = Column(Integer, nullable=False)
        release = Column(Text, nullable=False, index=True)
        provider = Column(Text, nullable=False)
        season = Column(Integer, nullable=False)
        episode = Column(Integer, nullable=False)
        old_status = Column(Enum(EpisodeStatus), nullable=False)

    class FailedSnatch(base):
        __tablename__ = 'failed_snatches'
        __table_args__ = (
            ForeignKeyConstraint(['series_id', 'series_provider_id'], ['tv_shows.series_id', 'tv_shows.series_provider_id'], ondelete='CASCADE',
                                 name=f'fk_{__tablename__}_series_id_series_provider_id'),
        )

        id = Column(Integer, autoincrement=True, primary_key=True)
        series_id = Column(Integer, nullable=False)
        series_provider_id = Column(Enum(SeriesProviderID), nullable=False)
        release = Column(Text, nullable=False, index=True)
        size = Column(Integer, nullable=False)
        provider = Column(Text, nullable=False)



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.



import re

from twilio.base.exceptions import TwilioRestException
from twilio.rest import TwilioRestClient

import sickrage
from sickrage.notification_providers import NotificationProvider


class AlexaNotification(NotificationProvider):
    def __init__(self):
        super(AlexaNotification, self).__init__()
        self.name = 'alexa'

    def notify_snatch(self, ep_name):
        if sickrage.app.config.alexa.notify_on_snatch:
            self._notify_alexa(self.notifyStrings[self.NOTIFY_SNATCH] + ': ' + ep_name)

    def notify_download(self, ep_name):
        if sickrage.app.config.alexa.notify_on_download:
            self._notify_alexa(self.notifyStrings[self.NOTIFY_DOWNLOAD] + ': ' + ep_name)

    def notify_subtitle_download(self, ep_name, lang):
        if sickrage.app.config.alexa.notify_on_subtitle_download:
            self._notify_alexa(self.notifyStrings[self.NOTIFY_SUBTITLE_DOWNLOAD] + ' ' + ep_name + ': ' + lang)

    def notify_version_update(self, new_version):
        if sickrage.app.config.alexa.enable:
            update_text = self.notifyStrings[self.NOTIFY_GIT_UPDATE_TEXT]
            self._notify_alexa(update_text + new_version)

    def notify_login(self, ipaddress=""):
        if sickrage.app.config.alexa.enable:
            update_text = self.notifyStrings[self.NOTIFY_LOGIN_TEXT]
            title = self.notifyStrings[self.NOTIFY_LOGIN]
            self._notify_alexa(title + " - " + update_text.format(ipaddress))

    def test_notify(self):
        if self._notify_alexa('This is a test notification', force=True):
            return True

    def _notify_alexa(self, message='', force=False):
        if not (sickrage.app.config.twilio.enable or force):
            return False

        sickrage.app.log.debug('Sending Alexa Notification: ' + message)
        if not sickrage.app.api.alexa.send_notification(message):
            sickrage.app.log.error('Alexa notification failed')
            return False

        return True



"""Initial migration

Revision ID: 8
Revises:
Create Date: 2017-12-29 14:39:27.854291

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '8'
down_revision = '7'


def upgrade():
    conn = op.get_bind()
    meta = sa.MetaData(bind=conn)

    if conn.engine.dialect.has_table(conn.engine, 'quicksearch_shows'):
        op.drop_table('quicksearch_shows')

    if conn.engine.dialect.has_table(conn.engine, 'quicksearch_episodes'):
        op.drop_table('quicksearch_episodes')



def downgrade():
    pass



"""Initial migration

Revision ID: 16
Revises:
Create Date: 2017-12-29 14:39:27.854291

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '16'
down_revision = '15'


def upgrade():
    conn = op.get_bind()
    meta = sa.MetaData(bind=conn)

    if conn.engine.dialect.has_table(conn.engine, 'scene_numbering'):
        scene_numbering = sa.Table('scene_numbering', meta, autoload=True)
        with op.get_context().begin_transaction():
            for row in conn.execute(scene_numbering.select()):
                conn.execute(
                    f'UPDATE tv_episodes SET scene_season = {row.scene_season} WHERE tv_episodes.showid = {row.indexer_id} and tv_episodes.season = {row.season} and tv_episodes.episode = {row.episode}')
                conn.execute(
                    f'UPDATE tv_episodes SET scene_episode = {row.scene_episode} WHERE tv_episodes.showid = {row.indexer_id} and tv_episodes.season = {row.season} and tv_episodes.episode = {row.episode}')

        op.drop_table('scene_numbering')


def downgrade():
    pass



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
# Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.


import locale

import sickrage
from sickrage.core.enums import TimezoneDisplay

date_presets = (
    '%Y-%m-%d',
    '%a, %Y-%m-%d',
    '%A, %Y-%m-%d',
    '%y-%m-%d',
    '%a, %y-%m-%d',
    '%A, %y-%m-%d',
    '%m/%d/%Y',
    '%a, %m/%d/%Y',
    '%A, %m/%d/%Y',
    '%m/%d/%y',
    '%a, %m/%d/%y',
    '%A, %m/%d/%y',
    '%m-%d-%Y',
    '%a, %m-%d-%Y',
    '%A, %m-%d-%Y',
    '%m-%d-%y',
    '%a, %m-%d-%y',
    '%A, %m-%d-%y',
    '%m.%d.%Y',
    '%a, %m.%d.%Y',
    '%A, %m.%d.%Y',
    '%m.%d.%y',
    '%a, %m.%d.%y',
    '%A, %m.%d.%y',
    '%d-%m-%Y',
    '%a, %d-%m-%Y',
    '%A, %d-%m-%Y',
    '%d-%m-%y',
    '%a, %d-%m-%y',
    '%A, %d-%m-%y',
    '%d/%m/%Y',
    '%a, %d/%m/%Y',
    '%A, %d/%m/%Y',
    '%d/%m/%y',
    '%a, %d/%m/%y',
    '%A, %d/%m/%y',
    '%d.%m.%Y',
    '%a, %d.%m.%Y',
    '%A, %d.%m.%Y',
    '%d.%m.%y',
    '%a, %d.%m.%y',
    '%A, %d.%m.%y',
    '%d. %b %Y',
    '%a, %d. %b %Y',
    '%A, %d. %b %Y',
    '%d. %b %y',
    '%a, %d. %b %y',
    '%A, %d. %b %y',
    '%d. %B %Y',
    '%a, %d. %B %Y',
    '%A, %d. %B %Y',
    '%d. %B %y',
    '%a, %d. %B %y',
    '%A, %d. %B %y',
    '%b %d, %Y',
    '%a, %b %d, %Y',
    '%A, %b %d, %Y',
    '%B %d, %Y',
    '%a, %B %d, %Y',
    '%A, %B %d, %Y'
)

time_presets = ('%I:%M:%S %p', '%H:%M:%S')


class SRDateTime(object):
    def __init__(self, dt, convert=False):
        self.dt = dt
        if convert and sickrage.app.config.gui.timezone_display == TimezoneDisplay.LOCAL:
            try:
                self.dt = dt.astimezone(sickrage.app.tz)
            except Exception as e:
                pass

        self.has_locale = True
        self.en_US_norm = locale.normalize('en_US.utf-8')

    # display Time in SickRage Format
    def srftime(self, show_seconds=False, t_preset=None):
        """
        Display time in SR format

        :param show_seconds: Boolean, show seconds
        :param t_preset: Preset time format
        :return: time string
        """

        strt = ''

        try:
            locale.setlocale(locale.LC_TIME, '')
        except Exception:
            pass

        try:
            if self.has_locale:
                locale.setlocale(locale.LC_TIME, locale.normalize(sickrage.app.config.gui.gui_lang))
        except Exception:
            try:
                if self.has_locale:
                    locale.setlocale(locale.LC_TIME, self.en_US_norm)
            except Exception:
                self.has_locale = False

        try:
            if t_preset is not None:
                strt = self.dt.strftime(t_preset)
            elif show_seconds:
                strt = self.dt.strftime(sickrage.app.config.gui.time_preset_w_seconds)
            else:
                strt = self.dt.strftime(sickrage.app.config.gui.time_preset)
        finally:
            try:
                if self.has_locale:
                    locale.setlocale(locale.LC_TIME, '')
            except Exception:
                self.has_locale = False

        return strt

    # display Date in SickRage Format
    def srfdate(self, d_preset=None):
        """
        Display date in SR format

        :param d_preset: Preset date format
        :return: date string
        """

        strd = ''

        try:
            locale.setlocale(locale.LC_TIME, '')
        except Exception:
            pass

        try:
            if self.has_locale:
                locale.setlocale(locale.LC_TIME, locale.normalize(sickrage.app.config.gui.gui_lang))
        except Exception:
            try:
                if self.has_locale:
                    locale.setlocale(locale.LC_TIME, self.en_US_norm)
            except Exception:
                self.has_locale = False

        try:
            if d_preset is not None:
                strd = self.dt.strftime(d_preset)
            else:
                strd = self.dt.strftime(sickrage.app.config.gui.date_preset)
        finally:
            try:
                locale.setlocale(locale.LC_TIME, '')
            except Exception:
                pass

        return strd

    # display Datetime in SickRage Format
    def srfdatetime(self, show_seconds=False, d_preset=None, t_preset=None):
        """
        Show datetime in SR format

        :param show_seconds: Boolean, show seconds as well
        :param d_preset: Preset date format
        :param t_preset: Preset time format
        :return: datetime string
        """

        strd = ''

        try:
            locale.setlocale(locale.LC_TIME, '')
        except Exception:
            pass

        try:
            if d_preset is not None:
                strd = self.dt.strftime(d_preset)
            else:
                strd = self.dt.strftime(sickrage.app.config.gui.date_preset)

            try:
                if self.has_locale:
                    locale.setlocale(locale.LC_TIME, locale.normalize(sickrage.app.config.gui.gui_lang))
            except Exception:
                try:
                    if self.has_locale:
                        locale.setlocale(locale.LC_TIME, self.en_US_norm)
                except Exception:
                    self.has_locale = False

            if t_preset is not None:
                strd += ', {}'.format(self.dt.strftime(t_preset))
            elif show_seconds:
                strd += ', {}'.format(self.dt.strftime(sickrage.app.config.gui.time_preset_w_seconds))
            else:
                strd += ', {}'.format(self.dt.strftime(sickrage.app.config.gui.time_preset))
        finally:
            try:
                if self.has_locale:
                    locale.setlocale(locale.LC_TIME, '')
            except Exception:
                self.has_locale = False

        return strd



import adba
import sickrage
from adba import AniDBCommandTimeoutError
from sickrage.core.exceptions import AnidbAdbaConnectionException


def set_up_anidb_connection():
    """Connect to anidb."""
    if not sickrage.app.config.anidb.enable:
        sickrage.app.log.debug('Usage of AniDB disabled. Skipping')
        return False

    if not sickrage.app.config.anidb.username and not sickrage.app.config.anidb.password:
        sickrage.app.log.debug('AniDB username and/or password are not set. Aborting anidb lookup.')
        return False

    if not sickrage.app.adba_connection:
        try:
            sickrage.app.adba_connection = adba.Connection(keepAlive=True)
        except Exception as error:
            sickrage.app.log.warning('AniDB exception msg: {0!r}'.format(error))
            return False

    try:
        if not sickrage.app.adba_connection.authed():
            sickrage.app.adba_connection.auth(sickrage.app.config.anidb.username, sickrage.app.config.anidb.password)
        else:
            return True
    except Exception as error:
        sickrage.app.log.warning('AniDB exception msg: {0!r}'.format(error))
        return False

    return sickrage.app.adba_connection.authed()


def get_release_groups_for_anime(series_name):
    """Get release groups for an anidb anime."""
    groups = []

    if set_up_anidb_connection():
        try:
            anime = adba.Anime(sickrage.app.adba_connection, name=series_name)
            groups = anime.get_groups()
        except Exception as error:
            sickrage.app.log.warning('Unable to retrieve Fansub Groups from AniDB. Error: {}'.format(error))
            raise AnidbAdbaConnectionException(error)

    return groups


def get_short_group_name(release_group):
    short_group_list = []

    try:
        group = sickrage.app.adba_connection.group(gname=release_group)
    except AniDBCommandTimeoutError:
        sickrage.app.log.debug('Timeout while loading group from AniDB. Trying next group')
    except Exception:
        sickrage.app.log.debug('Failed while loading group from AniDB. Trying next group')
    else:
        for line in group.datalines:
            if line['shortname']:
                short_group_list.append(line['shortname'])
            else:
                if release_group not in short_group_list:
                    short_group_list.append(release_group)

    return short_group_list


def short_group_names(groups):
    """
    Find AniDB short group names for release groups

    :param groups: list of groups to find short group names for
    :return: list of shortened group names
    """
    groups = groups.split(",")
    short_group_list = []

    if set_up_anidb_connection():
        for group_name in groups:
            short_group_list += get_short_group_name(group_name) or [group_name]
    else:
        short_group_list = groups

    return short_group_list


def get_anime_episode(file_path):
    """
    Look up anidb properties for an episode

    :param file_path: file to check
    :return: episode object
    """
    ep = None

    if set_up_anidb_connection():
        ep = adba.Episode(sickrage.app.adba_connection, filePath=file_path,
                          paramsF=[
                              "quality",
                              "anidb_file_name",
                              "crc32"
                          ],
                          paramsA=[
                              "epno",
                              "english_name",
                              "short_name_list",
                              "other_name",
                              "synonym_list"
                          ])

    return ep



# coding=utf-8
# Author: Gonçalo M. (aka duramato/supergonkas) <supergonkas@gmail.com>
#
# URL: https://sickrage.ca
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE. If not, see <http://www.gnu.org/licenses/>.


import sickrage
from sickrage.core.caches.tv_cache import TVCache
from sickrage.core.helpers import validate_url, try_int, convert_size
from sickrage.search_providers import TorrentProvider


class TorrentProjectProvider(TorrentProvider):
    def __init__(self):
        super(TorrentProjectProvider, self).__init__('TorrentProject', 'https://torrentproject.se', False)

        # custom settings
        self.custom_settings = {
            'custom_url': '',
            'minseed': 0,
            'minleech': 0
        }

        # Cache
        self.cache = TVCache(self, search_strings={'RSS': ['0day']})

    def search(self, search_strings, age=0, series_id=None, series_provider_id=None, season=None, episode=None, **kwargs):
        results = []

        search_params = {
            'out': 'json',
            'filter': 2101,
            'showmagnets': 'on',
            'num': 50
        }

        for mode in search_strings:
            sickrage.app.log.debug("Search Mode: {0}".format(mode))

            for search_string in search_strings[mode]:

                if mode != 'RSS':
                    sickrage.app.log.debug("Search string: {0}".format
                                           (search_string))

                search_params['s'] = search_string

                search_url = self.urls['base_url']
                if self.custom_settings['custom_url']:
                    if not validate_url(self.custom_settings['custom_url']):
                        sickrage.app.log.warning("Invalid custom url set, please check your settings")
                        return results
                    search_url = self.custom_settings['custom_url']

                resp = self.session.get(search_url, params=search_params)
                if not resp or not resp.content:
                    sickrage.app.log.debug("No data returned from provider")
                    continue

                try:
                    data = resp.json()
                except ValueError:
                    sickrage.app.log.debug("No data returned from provider")
                    continue

                results += self.parse(data, mode)

        return results

    def parse(self, data, mode, **kwargs):
        """
        Parse search results from data
        :param data: response data
        :param mode: search mode
        :return: search results
        """

        results = []

        if not (data and "total_found" in data and int(data["total_found"]) > 0):
            sickrage.app.log.debug("Data returned from provider does not contain any torrents")
            return results

        del data["total_found"]

        for i in data:
            try:
                title = data[i]["title"]
                download_url = data[i]["magnet"]
                if not all([title, download_url]):
                    continue

                seeders = try_int(data[i]["seeds"], 1)
                leechers = try_int(data[i]["leechs"], 0)
                torrent_size = data[i]["torrent_size"]

                size = convert_size(torrent_size, -1)

                results += [
                    {'title': title, 'link': download_url, 'size': size, 'seeders': seeders, 'leechers': leechers}
                ]

                if mode != 'RSS':
                    sickrage.app.log.debug("Found result: {}".format(title))
            except Exception:
                sickrage.app.log.error("Failed parsing provider.")

        return results



from __future__ import absolute_import, division, print_function

from trakt.core.helpers import from_iso8601_datetime, to_iso8601_datetime, \
    from_iso8601_date, to_iso8601_date, deprecated
from trakt.objects.core.helpers import update_attributes
from trakt.objects.video import Video


class Movie(Video):
    def __init__(self, client, keys, index=None):
        super(Movie, self).__init__(client, keys, index)

        self.title = None
        """
        :type: :class:`~python:str`

        Title
        """

        self.year = None
        """
        :type: :class:`~python:int`

        Year
        """

        self.watchers = None  # trending
        """
        :type: :class:`~python:int`

        Number of active watchers (returned by the :code:`Trakt['movies'].trending()`
        and :code:`Trakt['shows'].trending()` methods)
        """

        self.tagline = None
        """
        :type: :class:`~python:str`

        Tagline
        """

        self.released = None
        """
        :type: :class:`~python:datetime.date`

        Release date
        """

        self.runtime = None
        """
        :type: :class:`~python:int`

        Duration (in minutes)
        """

        self.certification = None
        """
        :type: :class:`~python:str`

        Content certification (e.g :code:`PG-13`)
        """

        self.updated_at = None
        """
        :type: :class:`~python:datetime.datetime`

        Updated date/time
        """

        self.homepage = None
        """
        :type: :class:`~python:str`

        Homepage URL
        """

        self.trailer = None
        """
        :type: :class:`~python:str`

        Trailer URL
        """

        self.language = None
        """
        :type: :class:`~python:str`

        Language (for title, overview, etc..)
        """

        self.available_translations = None
        """
        :type: :class:`~python:list`

        Available translations (for title, overview, etc..)
        """

        self.genres = None
        """
        :type: :class:`~python:list`

        Genres
        """

    def to_identifier(self):
        """Return the movie identifier which is compatible with requests that require movie definitions.

        :return: Movie identifier/definition
        :rtype: :class:`~python:dict`
        """

        return {
            'ids': dict(self.keys),
            'title': self.title,
            'year': self.year
        }

    @deprecated('Movie.to_info() has been moved to Movie.to_dict()')
    def to_info(self):
        """**Deprecated:** use the :code:`to_dict()` method instead."""
        return self.to_dict()

    def to_dict(self):
        """Dump movie to a dictionary.

        :return: Movie dictionary
        :rtype: :class:`~python:dict`
        """

        result = self.to_identifier()

        result.update({
            'watched': 1 if self.is_watched else 0,
            'collected': 1 if self.is_collected else 0,

            'plays': self.plays if self.plays is not None else 0,
            'in_watchlist': self.in_watchlist if self.in_watchlist is not None else 0,
            'progress': self.progress,

            'last_watched_at': to_iso8601_datetime(self.last_watched_at),
            'collected_at': to_iso8601_datetime(self.collected_at),
            'paused_at': to_iso8601_datetime(self.paused_at)
        })

        if self.rating:
            result['rating'] = self.rating.value
            result['rated_at'] = to_iso8601_datetime(self.rating.timestamp)

        # Extended Info
        if self.released:
            result['released'] = to_iso8601_date(self.released)

        if self.updated_at:
            result['updated_at'] = to_iso8601_datetime(self.updated_at)

        if self.overview:
            result['overview'] = self.overview

        if self.tagline:
            result['tagline'] = self.tagline

        if self.runtime:
            result['runtime'] = self.runtime

        if self.certification:
            result['certification'] = self.certification

        if self.homepage:
            result['homepage'] = self.homepage

        if self.trailer:
            result['trailer'] = self.trailer

        if self.language:
            result['language'] = self.language

        if self.available_translations:
            result['available_translations'] = self.available_translations

        if self.genres:
            result['genres'] = self.genres

        return result

    def _update(self, info=None, **kwargs):
        if not info:
            return

        super(Movie, self)._update(info, **kwargs)

        update_attributes(self, info, [
            'title',

            # Trending
            'watchers',

            # Extended Info
            'tagline',
            'certification',
            'homepage',
            'trailer',
            'language',
            'available_translations',
            'genres'
        ])

        # Ensure `year` attribute is an integer (fixes incorrect type returned by search)
        if info.get('year'):
            self.year = int(info['year'])

        # Extended Info
        if info.get('runtime'):
            self.runtime = info['runtime']

        if 'released' in info:
            self.released = from_iso8601_date(info.get('released'))

        if 'updated_at' in info:
            self.updated_at = from_iso8601_datetime(info.get('updated_at'))

    @classmethod
    def _construct(cls, client, keys, info, index=None, **kwargs):
        movie = cls(client, keys, index=index)
        movie._update(info, **kwargs)

        return movie

    def __repr__(self):
        return '<Movie %r (%s)>' % (self.title, self.year)



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################
import functools
import time
import traceback
import types
from concurrent.futures.thread import ThreadPoolExecutor
from typing import Optional, Awaitable
from urllib.parse import urlparse, urljoin

import bleach
from jose import ExpiredSignatureError
from keycloak.exceptions import KeycloakClientError
from mako.exceptions import RichTraceback
from tornado import locale
from tornado.escape import utf8
from tornado.ioloop import IOLoop
from tornado.web import RequestHandler

import sickrage
from sickrage.core.helpers import is_ip_whitelisted, torrent_webui_url


class BaseHandler(RequestHandler):
    def __init__(self, application, request, **kwargs):
        super(BaseHandler, self).__init__(application, request, **kwargs)

        self.executor = ThreadPoolExecutor(thread_name_prefix='WEB-Thread')

        self.startTime = time.time()

    def data_received(self, chunk: bytes) -> Optional[Awaitable[None]]:
        pass

    def get_user_locale(self):
        return locale.get(sickrage.app.config.gui.gui_lang)

    def write_error(self, status_code, **kwargs):
        if status_code not in [401, 404] and "exc_info" in kwargs:
            exc_info = kwargs["exc_info"]
            error = repr(exc_info[1])

            sickrage.app.log.error(error)

            if self.settings.get("debug"):
                trace_info = ''.join([f"{line}<br>" for line in traceback.format_exception(*exc_info)])
                request_info = ''.join([f"<strong>{k}</strong>: {v}<br>" for k, v in self.request.__dict__.items()])

                self.set_header('Content-Type', 'text/html')
                return self.finish(f"""<html>
                                 <title>{error}</title>
                                 <body>
                                    <button onclick="window.location='{sickrage.app.config.general.web_root}/logs/';">View Log(Errors)</button>
                                    <button onclick="window.location='{sickrage.app.config.general.web_root}/home/restart?pid={sickrage.app.pid}&force=1';">Restart SiCKRAGE</button>
                                    <button onclick="window.location='{sickrage.app.config.general.web_root}/logout';">Logout</button>
                                    <h2>Error</h2>
                                    <p>{error}</p>
                                    <h2>Traceback</h2>
                                    <p>{trace_info}</p>
                                    <h2>Request Info</h2>
                                    <p>{request_info}</p>
                                 </body>
                               </html>""")

    def get_current_user(self):
        if is_ip_whitelisted(self.request.remote_ip):
            return True
        elif sickrage.app.config.general.sso_auth_enabled and sickrage.app.auth_server.health:
            try:
                access_token = self.get_secure_cookie('_sr_access_token')
                refresh_token = self.get_secure_cookie('_sr_refresh_token')
                if not all([access_token, refresh_token]):
                    return

                certs = sickrage.app.auth_server.certs()
                if not certs:
                    return

                try:
                    return sickrage.app.auth_server.decode_token(access_token.decode("utf-8"), certs)
                except (KeycloakClientError, ExpiredSignatureError):
                    token = sickrage.app.auth_server.refresh_token(refresh_token.decode("utf-8"))
                    if not token:
                        return

                    self.set_secure_cookie('_sr_access_token', token['access_token'])
                    self.set_secure_cookie('_sr_refresh_token', token['refresh_token'])
                    return sickrage.app.auth_server.decode_token(token['access_token'], certs)
            except Exception as e:
                return
        elif sickrage.app.config.general.local_auth_enabled:
            cookie = self.get_secure_cookie('_sr').decode() if self.get_secure_cookie('_sr') else None
            if cookie == sickrage.app.config.general.api_v1_key:
                return True

    def render(self, template_name, **kwargs):
        template_kwargs = {
            'title': "",
            'header': "",
            'topmenu': "",
            'submenu': "",
            'controller': "home",
            'action': "index",
            'srPID': sickrage.app.pid,
            'srHttpsEnabled': sickrage.app.config.general.enable_https or bool(self.request.headers.get('X-Forwarded-Proto') == 'https'),
            'srHost': self.request.headers.get('X-Forwarded-Host', self.request.host.split(':')[0]),
            'srHttpPort': self.request.headers.get('X-Forwarded-Port', sickrage.app.config.general.web_port),
            'srHttpsPort': sickrage.app.config.general.web_port,
            'srHandleReverseProxy': sickrage.app.config.general.handle_reverse_proxy,
            'srDefaultPage': sickrage.app.config.general.default_page.value,
            'srWebRoot': sickrage.app.config.general.web_root,
            'srLocale': self.get_user_locale().code,
            'srLocaleDir': sickrage.LOCALE_DIR,
            'srStartTime': self.startTime,
            'makoStartTime': time.time(),
            'overall_stats': None,
            'torrent_webui_url': torrent_webui_url(),
            'application': self.application,
            'request': self.request,
        }

        template_kwargs.update(self.get_template_namespace())
        template_kwargs.update(kwargs)

        try:
            return self.application.settings['templates'][template_name].render_unicode(**template_kwargs)
        except Exception:
            kwargs['title'] = _('HTTP Error 500')
            kwargs['header'] = _('HTTP Error 500')
            kwargs['backtrace'] = RichTraceback()
            template_kwargs.update(kwargs)

            sickrage.app.log.error("%s: %s" % (str(kwargs['backtrace'].error.__class__.__name__), kwargs['backtrace'].error))

            return self.application.settings['templates']['errors/500.mako'].render_unicode(**template_kwargs)

    def set_default_headers(self):
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Headers", "Content-Type, Access-Control-Allow-Headers, Authorization, X-Requested-With")
        self.set_header('Access-Control-Allow-Methods', 'POST, GET, PUT, PATCH, DELETE, OPTIONS')
        self.set_header('Cache-Control', 'no-store, no-cache, must-revalidate, max-age=0')

    def redirect(self, url, permanent=True, status=None, add_web_root=True):
        if add_web_root and sickrage.app.config.general.web_root not in url:
            url = urljoin(sickrage.app.config.general.web_root + '/', url.lstrip('/'))

        if self._headers_written:
            raise Exception("Cannot redirect after headers have been written")
        if status is None:
            status = 301 if permanent else 302
        else:
            assert isinstance(status, int) and 300 <= status <= 399
        self.set_status(status)
        self.set_header("Location", utf8(url))

    def previous_url(self):
        url = urlparse(self.request.headers.get("referer", "/{}/".format(sickrage.app.config.general.default_page.value)))
        return url._replace(scheme="", netloc="").geturl()

    def _genericMessage(self, subject, message):
        return self.render('generic_message.mako',
                           message=message,
                           subject=subject,
                           title="",
                           controller='root',
                           action='genericmessage')

    def get_url(self, url):
        if sickrage.app.config.general.web_root not in url:
            url = urljoin(sickrage.app.config.general.web_root + '/', url.lstrip('/'))
        url = urljoin("{}://{}".format(self.request.protocol, self.request.host), url)
        return url

    def run_async(self, method):
        @functools.wraps(method)
        async def wrapper(self, *args, **kwargs):
            resp = await IOLoop.current().run_in_executor(self.executor, functools.partial(method, *args, **kwargs))
            self.finish(resp)

        return types.MethodType(wrapper, self)

    def prepare(self):
        method_name = self.request.method.lower()
        method = self.run_async(getattr(self, method_name))
        setattr(self, method_name, method)

    def options(self, *args, **kwargs):
        self.set_status(204)
        self.finish()

    def get_argument(self, *args, **kwargs):
        value = super(BaseHandler, self).get_argument(*args, **kwargs)

        try:
            return bleach.clean(value)
        except TypeError:
            return value



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################
import re

from sqlalchemy import orm

import sickrage
from sickrage.core.databases.main import MainDB
from sickrage.core.enums import SeriesProviderID
from sickrage.core.tv.show.helpers import find_show


def map_series_providers(series_provider_id, series_id, name):
    session = sickrage.app.main_db.session()

    mapped = {}
    for series_provider_id in SeriesProviderID:
        mapped[series_provider_id.name] = None

    # init mapped series_provider_ids object
    for mapped_series_provider_id in SeriesProviderID:
        if mapped_series_provider_id == series_provider_id:
            mapped[mapped_series_provider_id.name] = series_id

    # for each mapped entry
    for dbData in session.query(MainDB.SeriesProviderMapping).filter_by(series_id=series_id, series_provider_id=series_provider_id):
        # Check if its mapped with both tvdb and tvrage.
        if len([i for i in dbData if i is not None]) >= 4:
            sickrage.app.log.debug("Found series_provider_id mapping in cache for show: " + name)
            mapped[dbData.mapped_series_provider_id.name] = dbData.mapped_series_id
            return mapped
    else:
        for mapped_series_provider_id in SeriesProviderID:
            if mapped_series_provider_id == series_provider_id:
                mapped[mapped_series_provider_id.name] = series_id
                continue

            mapped_series_provider = sickrage.app.series_provider[mapped_series_provider_id]

            mapped_show = mapped_series_provider.search(name)
            if not mapped_show:
                continue

            if mapped_show and len(mapped_show) == 1:
                sickrage.app.log.debug(f"Mapping {sickrage.app.series_providers[series_provider_id].name} -> {mapped_series_provider} for show: {name}")

                mapped[mapped_series_provider_id.name] = int(mapped_show['id'])

                sickrage.app.log.debug("Adding series_provider_id mapping to DB for show: " + name)

                try:
                    session.query(MainDB.SeriesProviderMapping).filter_by(series_id=series_id, series_provider_id=series_provider_id,
                                                                          mapped_series_id=int(mapped_show['id'])).one()
                except orm.exc.NoResultFound:
                    session.add(MainDB.SeriesProviderMapping(**{
                        'series_id': series_id,
                        'series_provider_id': series_provider_id,
                        'mapped_series_id': int(mapped_show['id']),
                        'mapped_series_provider_id': mapped_series_provider_id['id']
                    }))
                    session.commit()

    return mapped


def search_series_provider_for_series_id(show_name, series_provider_id):
    """
    Contacts series provider to check for information on shows by series name to retrieve series id

    :param show_name: Name of show
    :param series_provider_id: series provider id
    :return:
    """

    show_name = re.sub('[. -]', ' ', show_name)

    series_provider = sickrage.app.series_providers[series_provider_id]

    # Query series provider for search term and build the list of results
    sickrage.app.log.debug("Trying to find show ID for show {} on series provider {}".format(show_name, series_provider.name))

    series_provider_data = series_provider.search(show_name)
    if not series_provider_data:
        return

    # try to pick a show that's in my show list
    for series in series_provider_data:
        series_id = series.get('id', None)
        if not series_id:
            continue

        if find_show(int(series_id), series_provider_id):
            return series_id

    return series_provider_data[0].get('id', None)



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.


import importlib
import inspect
import os
import pkgutil
import re
from xml.etree.ElementTree import ElementTree

import fanart

import sickrage
from sickrage.core.enums import SeriesProviderID
from sickrage.core.helpers import chmod_as_parent, replace_extension, try_int
from sickrage.core.websession import WebSession
from sickrage.series_providers.exceptions import SeriesProviderEpisodeNotFound, SeriesProviderSeasonNotFound
from sickrage.series_providers.helpers import map_series_providers


class MetadataProvider(object):
    """
    Base class for all metadata providers. Default behavior is meant to mostly
    follow KODI 12+ metadata standards. Has support for:
    - show metadata file
    - episode metadata file
    - episode thumbnail
    - show fanart
    - show poster
    - show banner
    - season thumbnails (poster)
    - season thumbnails (banner)
    - season all poster
    - season all banner
    """

    def __init__(self,
                 show_metadata=False,
                 episode_metadata=False,
                 fanart=False,
                 poster=False,
                 banner=False,
                 episode_thumbnails=False,
                 season_posters=False,
                 season_banners=False,
                 season_all_poster=False,
                 season_all_banner=False,
                 enabled=False):

        self.name = "Generic"
        self.enabled = enabled

        self._ep_nfo_extension = "nfo"
        self._show_metadata_filename = "tvshow.nfo"

        self.fanart_name = "fanart.jpg"
        self.poster_name = "poster.jpg"
        self.banner_name = "banner.jpg"

        self.season_all_poster_name = "season-all-poster.jpg"
        self.season_all_banner_name = "season-all-banner.jpg"

        self.show_metadata = show_metadata
        self.episode_metadata = episode_metadata
        self.fanart = fanart
        self.poster = poster
        self.banner = banner
        self.episode_thumbnails = episode_thumbnails
        self.season_posters = season_posters
        self.season_banners = season_banners
        self.season_all_poster = season_all_poster
        self.season_all_banner = season_all_banner

    @property
    def id(self):
        return str(re.sub(r"[^\w\d_]", "_", str(re.sub(r"[+]", "plus", self.name))).lower())

    @property
    def config(self):
        return "|".join(map(str, map(int, [self.show_metadata, self.episode_metadata, self.fanart, self.poster, self.banner, self.episode_thumbnails,
                                           self.season_posters, self.season_banners, self.season_all_poster, self.season_all_banner, self.enabled])))

    @config.setter
    def config(self, values):
        if not values:
            values = '0|0|0|0|0|0|0|0|0|0|0'

        self.show_metadata, self.episode_metadata, self.fanart, self.poster, self.banner, self.episode_thumbnails, self.season_posters, \
        self.season_banners, self.season_all_poster, self.season_all_banner, self.enabled = tuple(map(bool, map(int, values.split('|'))))

    @staticmethod
    def _check_exists(location):
        if location:
            result = os.path.isfile(location)
            return result
        return False

    def _has_show_metadata(self, show_obj):
        return self._check_exists(self.get_show_file_path(show_obj))

    def _has_episode_metadata(self, ep_obj):
        return self._check_exists(self.get_episode_file_path(ep_obj))

    def _has_fanart(self, show_obj):
        return self._check_exists(self.get_fanart_path(show_obj))

    def _has_poster(self, show_obj):
        return self._check_exists(self.get_poster_path(show_obj))

    def _has_banner(self, show_obj):
        return self._check_exists(self.get_banner_path(show_obj))

    def _has_episode_thumb(self, ep_obj):
        return self._check_exists(self.get_episode_thumb_path(ep_obj))

    def _has_season_poster(self, show_obj, season):
        return self._check_exists(self.get_season_poster_path(show_obj, season))

    def _has_season_banner(self, show_obj, season):
        return self._check_exists(self.get_season_banner_path(show_obj, season))

    def _has_season_all_poster(self, show_obj):
        return self._check_exists(self.get_season_all_poster_path(show_obj))

    def _has_season_all_banner(self, show_obj):
        return self._check_exists(self.get_season_all_banner_path(show_obj))

    def get_show_file_path(self, show_obj):
        return os.path.join(show_obj.location, self._show_metadata_filename)

    def get_episode_file_path(self, ep_obj):
        return replace_extension(ep_obj.location, self._ep_nfo_extension)

    def get_fanart_path(self, show_obj):
        return os.path.join(show_obj.location, self.fanart_name)

    def get_poster_path(self, show_obj):
        return os.path.join(show_obj.location, self.poster_name)

    def get_banner_path(self, show_obj):
        return os.path.join(show_obj.location, self.banner_name)

    @staticmethod
    def get_episode_thumb_path(ep_obj):
        """
        Returns the path where the episode thumbnail should be stored.
        ep_obj: a TVEpisode instance for which to create the thumbnail
        """
        if os.path.isfile(ep_obj.location):

            tbn_filename = ep_obj.location.rpartition(".")

            if tbn_filename[0] == "":
                tbn_filename = ep_obj.location + "-thumb.jpg"
            else:
                tbn_filename = tbn_filename[0] + "-thumb.jpg"
        else:
            return None

        return tbn_filename

    @staticmethod
    def get_season_poster_path(show_obj, season):
        """
        Returns the full path to the file for a given season poster.

        show_obj: a TVShow instance for which to generate the path
        season: a season number to be used for the path. Note that season 0
                means specials.
        """

        # Our specials thumbnail is, well, special
        if season == 0:
            season_poster_filename = 'season-specials'
        else:
            season_poster_filename = 'season' + str(season).zfill(2)

        return os.path.join(show_obj.location, season_poster_filename + '-poster.jpg')

    @staticmethod
    def get_season_banner_path(show_obj, season):
        """
        Returns the full path to the file for a given season banner.

        show_obj: a TVShow instance for which to generate the path
        season: a season number to be used for the path. Note that season 0
                means specials.
        """

        # Our specials thumbnail is, well, special
        if season == 0:
            season_banner_filename = 'season-specials'
        else:
            season_banner_filename = 'season' + str(season).zfill(2)

        return os.path.join(show_obj.location, season_banner_filename + '-banner.jpg')

    def get_season_all_poster_path(self, show_obj):
        return os.path.join(show_obj.location, self.season_all_poster_name)

    def get_season_all_banner_path(self, show_obj):
        return os.path.join(show_obj.location, self.season_all_banner_name)

    def _show_data(self, show_obj):
        """
        This should be overridden by the implementing class. It should
        provide the content of the show metadata file.
        """
        return None

    def _ep_data(self, ep_obj):
        """
        This should be overridden by the implementing class. It should
        provide the content of the episode metadata file.
        """
        return None

    def create_show_metadata(self, show_obj, force=False):
        if self.show_metadata and show_obj and (not self._has_show_metadata(show_obj) or force):
            sickrage.app.log.debug("Metadata provider " + self.name + " creating show metadata for " + show_obj.name)
            return self.write_show_file(show_obj)
        return False

    def create_episode_metadata(self, ep_obj, force=False):
        if self.episode_metadata and ep_obj and (not self._has_episode_metadata(ep_obj) or force):
            sickrage.app.log.debug("Metadata provider " + self.name + " creating episode metadata for " + ep_obj.pretty_name())
            return self.write_ep_file(ep_obj)
        return False

    def create_fanart(self, show_obj, which=0, force=False):
        if self.fanart and show_obj and (not self._has_fanart(show_obj) or force):
            sickrage.app.log.debug("Metadata provider " + self.name + " creating fanart for " + show_obj.name)
            return self.save_fanart(show_obj, which)
        return False

    def create_poster(self, show_obj, which=0, force=False):
        if self.poster and show_obj and (not self._has_poster(show_obj) or force):
            sickrage.app.log.debug("Metadata provider " + self.name + " creating poster for " + show_obj.name)
            return self.save_poster(show_obj, which)
        return False

    def create_banner(self, show_obj, which=0, force=False):
        if self.banner and show_obj and (not self._has_banner(show_obj) or force):
            sickrage.app.log.debug("Metadata provider " + self.name + " creating banner for " + show_obj.name)
            return self.save_banner(show_obj, which)
        return False

    def create_episode_thumb(self, ep_obj, force=False):
        if self.episode_thumbnails and ep_obj and (not self._has_episode_thumb(ep_obj) or force):
            sickrage.app.log.debug("Metadata provider " + self.name + " creating episode thumbnail for " + ep_obj.pretty_name())
            return self.save_thumbnail(ep_obj)
        return False

    def create_season_posters(self, show_obj, force=False):
        if self.season_posters and show_obj:
            result = []
            for ep_obj in show_obj.episodes:
                if not self._has_season_poster(show_obj, ep_obj.season) or force:
                    sickrage.app.log.debug("Metadata provider " + self.name + " creating season posters for " + show_obj.name)
                    result = result + [self.save_season_poster(show_obj, ep_obj.season)]
            return all(result)
        return False

    def create_season_banners(self, show_obj, force=False):
        if self.season_banners and show_obj:
            result = []
            sickrage.app.log.debug("Metadata provider " + self.name + " creating season banners for " + show_obj.name)
            for ep_obj in show_obj.episodes:
                if not self._has_season_banner(show_obj, ep_obj.season) or force:
                    result = result + [self.save_season_banner(show_obj, ep_obj.season)]
            return all(result)
        return False

    def create_season_all_poster(self, show_obj, force=False):
        if self.season_all_poster and show_obj and (not self._has_season_all_poster(show_obj) or force):
            sickrage.app.log.debug("Metadata provider " + self.name + " creating season all poster for " + show_obj.name)
            return self.save_season_all_poster(show_obj)
        return False

    def create_season_all_banner(self, show_obj, force=False):
        if self.season_all_banner and show_obj and (not self._has_season_all_banner(show_obj) or force):
            sickrage.app.log.debug("Metadata provider " + self.name + " creating season all banner for " + show_obj.name)
            return self.save_season_all_banner(show_obj)
        return False

    def _get_episode_thumb_url(self, ep_obj):
        """
        Returns the URL to use for downloading an episode's thumbnail. Uses
        theTVDB.com data.

        ep_obj: a TVEpisode object for which to grab the thumb URL
        """
        all_eps = [ep_obj] + ep_obj.related_episodes

        # validate show
        if not self.validate_show(ep_obj.show):
            return None

        # try all included episodes in case some have thumbs and others don't
        for cur_ep in all_eps:
            myEp = self.validate_show(cur_ep.show, cur_ep.season, cur_ep.episode)
            if not myEp:
                continue

            thumb_url = getattr(myEp, 'filename', None)
            if thumb_url:
                return thumb_url

        return None

    def write_show_file(self, show_obj):
        """
        Generates and writes show_obj's metadata under the given path to the
        filename given by get_show_file_path()

        show_obj: TVShow object for which to create the metadata

        path: An absolute or relative path where we should put the file. Note that
                the file name will be the default show_file_name.

        Note that this method expects that _show_data will return an ElementTree
        object. If your _show_data returns data in another format yo'll need to
        override this method.
        """

        data = self._show_data(show_obj)
        if not data:
            return False

        nfo_file_path = self.get_show_file_path(show_obj)
        nfo_file_dir = os.path.dirname(nfo_file_path)

        try:
            if not os.path.isdir(nfo_file_dir):
                sickrage.app.log.debug("Metadata dir didn't exist, creating it at " + nfo_file_dir)
                os.makedirs(nfo_file_dir)
                chmod_as_parent(nfo_file_dir)

            sickrage.app.log.debug("Writing show nfo file to " + nfo_file_path)

            with open(nfo_file_path, 'wb') as nfo_file:
                data.write(nfo_file, encoding='utf-8')

            chmod_as_parent(nfo_file_path)
        except IOError as e:
            sickrage.app.log.warning("Unable to write file to " + nfo_file_path + " - are you sure the folder is writable? {}".format(e))
            return False

        return True

    def write_ep_file(self, ep_obj):
        """
        Generates and writes ep_obj's metadata under the given path with the
        given filename root. Uses the episode's name with the extension in
        _ep_nfo_extension.

        ep_obj: TVEpisode object for which to create the metadata

        file_name_path: The file name to use for this metadata. Note that the extension
                will be automatically added based on _ep_nfo_extension. This should
                include an absolute path.

        Note that this method expects that _ep_data will return an ElementTree
        object. If your _ep_data returns data in another format yo'll need to
        override this method.
        """

        data = self._ep_data(ep_obj)
        if not data:
            return False

        nfo_file_path = self.get_episode_file_path(ep_obj)
        nfo_file_dir = os.path.dirname(nfo_file_path)

        try:
            if not os.path.isdir(nfo_file_dir):
                sickrage.app.log.debug("Metadata dir didn't exist, creating it at " + nfo_file_dir)
                os.makedirs(nfo_file_dir)
                chmod_as_parent(nfo_file_dir)

            sickrage.app.log.debug("Writing episode nfo file to " + nfo_file_path)

            with open(nfo_file_path, 'wb') as nfo_file:
                data.write(nfo_file, encoding='utf-8')

            chmod_as_parent(nfo_file_path)
        except IOError as e:
            sickrage.app.log.warning("Unable to write file to " + nfo_file_path + " - are you sure the folder is writable? {}".format(e))
            return False

        return True

    def save_thumbnail(self, ep_obj):
        """
        Retrieves a thumbnail and saves it to the correct spot. This method should not need to
        be overridden by implementing classes, changing get_episode_thumb_path and
        _get_episode_thumb_url should suffice.

        ep_obj: a TVEpisode object for which to generate a thumbnail
        """

        file_path = self.get_episode_thumb_path(ep_obj)

        if not file_path:
            sickrage.app.log.debug("Unable to find a file path to use for this thumbnail, not generating it")
            return False

        thumb_url = self._get_episode_thumb_url(ep_obj)

        # if we can't find one then give up
        if not thumb_url:
            sickrage.app.log.debug("No thumb is available for this episode, not creating a thumb")
            return False

        thumb_data = self.get_show_image(thumb_url)

        result = self._write_image(thumb_data, file_path)

        if not result:
            return False

        for cur_ep in [ep_obj] + ep_obj.related_episodes:
            cur_ep.hastbn = True
            cur_ep.save()

        return True

    def save_fanart(self, show_obj, which=0):
        """
        Downloads a fanart image and saves it to the filename specified by fanart_name
        inside the show's root folder.

        show_obj: a TVShow object for which to download fanart
        """

        # use the default fanart name
        fanart_path = self.get_fanart_path(show_obj)

        fanart_data = self._retrieve_show_image('fanart', show_obj, which)

        if not fanart_data:
            sickrage.app.log.debug("No fanart image was retrieved, unable to write fanart")
            return False

        return self._write_image(fanart_data, fanart_path)

    def save_poster(self, show_obj, which=0):
        """
        Downloads a poster image and saves it to the filename specified by poster_name
        inside the show's root folder.

        show_obj: a TVShow object for which to download a poster
        """

        # use the default poster name
        poster_path = self.get_poster_path(show_obj)

        poster_data = self._retrieve_show_image('poster', show_obj, which)

        if not poster_data:
            sickrage.app.log.debug("No show poster image was retrieved, unable to write poster")
            return False

        return self._write_image(poster_data, poster_path)

    def save_banner(self, show_obj, which=0):
        """
        Downloads a banner image and saves it to the filename specified by banner_name
        inside the show's root folder.

        show_obj: a TVShow object for which to download a banner
        """

        # use the default banner name
        banner_path = self.get_banner_path(show_obj)

        banner_data = self._retrieve_show_image('series', show_obj, which)

        if not banner_data:
            sickrage.app.log.debug("No show banner image was retrieved, unable to write banner")
            return False

        return self._write_image(banner_data, banner_path)

    def save_season_poster(self, show_obj, season, which=0):
        season_url = self._retrieve_season_poster_image(show_obj, season, which)

        season_poster_file_path = self.get_season_poster_path(show_obj, season)
        if not season_poster_file_path:
            sickrage.app.log.debug("Path for season " + str(season) + " came back blank, skipping this season")
            return False

        seasonData = self.get_show_image(season_url)
        if not seasonData:
            sickrage.app.log.debug("No season poster data available, skipping this season")
            return False

        return self._write_image(seasonData, season_poster_file_path)

    def save_season_banner(self, show_obj, season, which=0):
        season_url = self._retrieve_season_banner_image(show_obj, season, which)

        season_banner_file_path = self.get_season_banner_path(show_obj, season)
        if not season_banner_file_path:
            sickrage.app.log.debug("Path for season " + str(season) + " came back blank, skipping this season")
            return False

        seasonData = self.get_show_image(season_url)
        if not seasonData:
            sickrage.app.log.debug("No season banner data available, skipping this season")
            return False

        return self._write_image(seasonData, season_banner_file_path)

    def save_season_all_poster(self, show_obj, which=0):
        # use the default season all poster name
        poster_path = self.get_season_all_poster_path(show_obj)

        poster_data = self._retrieve_show_image('poster', show_obj, which)

        if not poster_data:
            sickrage.app.log.debug("No show poster image was retrieved, unable to write season all poster")
            return False

        return self._write_image(poster_data, poster_path)

    def save_season_all_banner(self, show_obj, which=0):
        # use the default season all banner name
        banner_path = self.get_season_all_banner_path(show_obj)

        banner_data = self._retrieve_show_image('series', show_obj, which)

        if not banner_data:
            sickrage.app.log.debug("No show banner image was retrieved, unable to write season all banner")
            return False

        return self._write_image(banner_data, banner_path)

    def _write_image(self, image_data, image_path, force=False):
        """
        Saves the data in image_data to the location image_path. Returns True/False
        to represent success or failure.

        image_data: binary image data to write to file
        image_path: file location to save the image to
        """

        # don't bother overwriting it
        if os.path.isfile(image_path) and not force:
            sickrage.app.log.debug("Image already exists, not downloading")
            return False

        image_dir = os.path.dirname(image_path)

        if not image_data:
            sickrage.app.log.debug("Unable to retrieve image to save in %s, skipping" % image_path)
            return False

        try:
            if not os.path.isdir(image_dir):
                sickrage.app.log.debug("Metadata dir didn't exist, creating it at " + image_dir)
                os.makedirs(image_dir)
                chmod_as_parent(image_dir)

            with open(image_path, 'wb') as outFile:
                outFile.write(image_data)

            chmod_as_parent(image_path)
        except IOError as e:
            sickrage.app.log.warning("Unable to write image to " + image_path + " - are you sure the show folder is writable? {}".format(e))
            return False

        return True

    def _retrieve_show_image(self, image_type, show_obj, which=0):
        """
        Gets an image URL from theTVDB.com and fanart.tv, downloads it and returns the data.

        image_type: type of image to retrieve (currently supported: fanart, poster, banner)
        show_obj: a TVShow object to use when searching for the image
        which: optional, a specific numbered poster to look for

        Returns: the binary image data if available, or else None
        """

        image_data = None

        if image_type not in ('fanart', 'poster', 'series', 'poster_thumb', 'series_thumb', 'fanart_thumb'):
            sickrage.app.log.warning(
                "Invalid image type " + str(image_type) + ", couldn't find it in the " + show_obj.series_provider.name + " object")
            return

        series_provider_language = show_obj.lang or sickrage.app.config.general.series_provider_default_language

        is_image_thumb = '_thumb' in image_type
        image_types = {
            '{}'.format(image_type): {
                'series_provider': lambda:
                show_obj.series_provider.images(show_obj.series_id, language=series_provider_language, key_type=image_type.replace('_thumb', ''))[which][
                    ('filename', 'thumbnail')[is_image_thumb]],
                'fanart': lambda: self._retrieve_show_images_from_fanart(show_obj, image_type.replace('_thumb', ''), is_image_thumb)
            }
        }

        for fname in ['series_provider', 'fanart']:
            try:
                image_url = image_types[image_type][fname]()
                if image_url:
                    image_data = self.get_show_image(image_url)
                    if image_data:
                        break
            except (KeyError, IndexError, TypeError) as e:
                pass

        return image_data

    @staticmethod
    def _retrieve_season_poster_image(show_obj, season, which=0):
        """
        Should return a dict like:

        result = {<season number>:
                    {1: '<url 1>', 2: <url 2>, ...},}
        """

        try:
            series_provider_language = show_obj.lang or sickrage.app.config.general.series_provider_default_language

            # Give us just the normal poster-style season graphics
            image_data = show_obj.series_provider.images(show_obj.series_id, language=series_provider_language, key_type='season', season=season)
            if image_data:
                return image_data[which]['filename']

            sickrage.app.log.debug("{}: No season {} poster images on {} to download found".format(show_obj.series_id, season, show_obj.series_provider.name))
        except (KeyError, IndexError):
            pass

    @staticmethod
    def _retrieve_season_banner_image(show_obj, season, which=0):
        """
        Should return a dict like:

        result = {<season number>:
                    {1: '<url 1>', 2: <url 2>, ...},}
        """

        try:
            series_provider_language = show_obj.lang or sickrage.app.config.general.series_provider_default_language

            # Give us just the normal season graphics
            image_data = show_obj.series_provider.images(show_obj.series_id, language=series_provider_language, key_type='seasonwide', season=season)
            if image_data:
                return image_data[which]['filename']

            sickrage.app.log.debug("{}: No season {} banner images on {} to download found".format(show_obj.series_id, season, show_obj.series_provider.name))
        except (KeyError, IndexError):
            pass

    def retrieve_show_metadata(self, folder) -> (int, str, int):
        """
        :param folder:
        :return:
        """

        empty_return = (None, None, None)

        metadata_path = os.path.join(folder, self._show_metadata_filename)

        if not os.path.isdir(folder) or not os.path.isfile(metadata_path):
            sickrage.app.log.debug("Can't load the metadata file from " + metadata_path + ", it doesn't exist")
            return empty_return

        try:
            sickrage.app.log.debug(f"Loading show info from SiCKRAGE metadata file in {folder}")

            with open(metadata_path, 'rb') as xmlFileObj:
                showXML = ElementTree(file=xmlFileObj)

            if showXML.findtext('title') is None or (
                    showXML.findtext('tvdbid') is None and showXML.findtext('id') is None):
                sickrage.app.log.info("Invalid info in tvshow.nfo (missing name or id): {} {} {}".format(showXML.findtext('title'),
                                                                                                         showXML.findtext('tvdbid'),
                                                                                                         showXML.findtext('id')))
                return empty_return

            name = showXML.findtext('title')

            series_id_text = showXML.findtext('tvdbid') or showXML.findtext('id')
            if series_id_text:
                series_id = try_int(series_id_text, None)
                if not series_id:
                    sickrage.app.log.debug("Invalid series id (" + str(series_id) + "), not using metadata file")
                    return empty_return
            else:
                sickrage.app.log.debug("Empty <id> or <tvdbid> field in NFO, unable to find a ID, not using metadata file")
                return empty_return

            if showXML.findtext('tvdbid') is not None:
                series_id = int(showXML.findtext('tvdbid'))
            elif showXML.findtext('id') is not None:
                series_id = int(showXML.findtext('id'))
            else:
                sickrage.app.log.warning("Empty <id> or <tvdbid> field in NFO, unable to find a ID")
                return empty_return

            epg_url_text = showXML.findtext('episodeguide/url')
            if epg_url_text:
                epg_url = epg_url_text.lower()
                if str(series_id) in epg_url and 'tvrage' in epg_url:
                    sickrage.app.log.warning("Invalid series id (" + str(series_id) + "), not using metadata file because it has TVRage info")
                    return empty_return

        except Exception as e:
            sickrage.app.log.warning("There was an error parsing your existing metadata file: '" + metadata_path + "' error: {}".format(e))
            return empty_return

        return series_id, name, SeriesProviderID.THETVDB

    @staticmethod
    def _retrieve_show_images_from_fanart(show, img_type, thumb=False):
        types = {
            'poster': fanart.TYPE.TV.POSTER,
            'series': fanart.TYPE.TV.BANNER,
            'fanart': fanart.TYPE.TV.BACKGROUND,
        }

        sickrage.app.log.debug("Searching for any " + img_type + " images on Fanart.tv for " + show.name)

        try:
            series_id = map_series_providers(show.series_provider_id, show.series_id, show.name)[SeriesProviderID.THETVDB.name]
            if series_id:
                request = fanart.Request(
                    apikey=sickrage.app.fanart_api_key,
                    id=series_id,
                    ws=fanart.WS.TV,
                    type=types[img_type],
                    sort=fanart.SORT.POPULAR,
                    limit=fanart.LIMIT.ONE,
                )

                resp = request.response()
                url = resp[types[img_type]][0]['url']
                if thumb:
                    url = re.sub('/fanart/', '/preview/', url)
                return url
        except Exception:
            pass

        sickrage.app.log.debug("Could not find any " + img_type + " images on Fanart.tv for " + show.name)

    @staticmethod
    def validate_show(show, season=None, episode=None):
        if season is None and episode is None:
            return

        try:
            series_provider_language = show.lang or sickrage.app.config.general.series_provider_default_language

            return show.series_provider.search(show.series_id, language=series_provider_language)[season][episode]
        except (SeriesProviderEpisodeNotFound, SeriesProviderSeasonNotFound):
            pass

    @staticmethod
    def get_show_image(url):
        if url is None:
            return None

        sickrage.app.log.debug("Fetching image from " + url)

        try:
            return WebSession().get(url, verify=False).content
        except Exception:
            sickrage.app.log.debug("There was an error trying to retrieve the image, aborting")


class MetadataProviders(dict):
    def __init__(self):
        super(MetadataProviders, self).__init__()
        for (__, name, __) in pkgutil.iter_modules([os.path.dirname(__file__)]):
            imported_module = importlib.import_module('.' + name, package='sickrage.metadata_providers')
            for __, klass in inspect.getmembers(imported_module,
                                                predicate=lambda o: inspect.isclass(o) and issubclass(o, MetadataProvider) and o is not MetadataProvider):
                self[klass().id] = klass()
                break



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################


import os
import re
from base64 import b64encode

import requests

import sickrage
from sickrage.clients import TorrentClient


class TransmissionAPI(TorrentClient):
    def __init__(self, host=None, username=None, password=None):
        super(TransmissionAPI, self).__init__('Transmission', host, username, password)

        if not self.host.endswith('/'):
            self.host += '/'

        if self.rpcurl.startswith('/'):
            self.rpcurl = self.rpcurl[1:]

        if self.rpcurl.endswith('/'):
            self.rpcurl = self.rpcurl[:-1]

        self.url = self.host + self.rpcurl + '/rpc'

    def _get_auth(self):
        self.response = self.session.post(self.url,
                                          timeout=120,
                                          json={'method': 'session-get'},
                                          auth=(self.username, self.password),
                                          verify=bool(sickrage.app.config.torrent.verify_cert))

        if self.response is not None and self.response.text:
            auth_match = re.search(r'X-Transmission-Session-Id:\s*(\w+)', self.response.text)
            if auth_match:
                self.auth = auth_match.group(1)
                self.session.headers.update({'x-transmission-session-id': self.auth})

                # Validating Transmission authorization
                success = self._request(method='post',
                                        json={'arguments': {}, 'method': 'session-get'},
                                        headers={'x-transmission-session-id': self.auth})

                if not success:
                    self.auth = None
            else:
                self.auth = None

        return self.auth

    def _add_torrent_uri(self, result):
        arguments = {
            'filename': result.url,
            'paused': 1 if sickrage.app.config.torrent.paused else 0,
        }

        if os.path.isabs(sickrage.app.config.torrent.path):
            arguments['download-dir'] = sickrage.app.config.torrent.path

        post_data = {
            'arguments': arguments,
            'method': 'torrent-add'
        }

        if self._request(method='post', json=post_data):
            return self.response.json()['result'] == "success"

    def _add_torrent_file(self, result):
        arguments = {
            'metainfo': b64encode(result.content),
            'paused': 1 if sickrage.app.config.torrent.paused else 0
        }

        if os.path.isabs(sickrage.app.config.torrent.path):
            arguments['download-dir'] = sickrage.app.config.torrent.path

        post_data = {
            'arguments': arguments,
            'method': 'torrent-add'
        }

        if self._request(method='post', json=post_data):
            return self.response.json()['result'] == "success"

    def _set_torrent_ratio(self, result):
        ratio = None
        if isinstance(result.ratio, int):
            ratio = result.ratio

        mode = 0
        if ratio:
            if float(ratio) == -1:
                ratio = 0
                mode = 2
            elif float(ratio) >= 0:
                ratio = float(ratio)
                mode = 1  # Stop seeding at seedRatioLimit

        arguments = {
            'ids': [result.hash],
            'seedRatioLimit': ratio,
            'seedRatioMode': mode
        }

        post_data = {
            'arguments': arguments,
            'method': 'torrent-set'
        }

        if self._request(method='post', json=post_data):
            return self.response.json()['result'] == "success"

    def _set_torrent_seed_time(self, result):
        if sickrage.app.config.torrent.seed_time and sickrage.app.config.torrent.seed_time != -1:
            time = int(60 * float(sickrage.app.config.torrent.seed_time))
            arguments = {
                'ids': [result.hash],
                'seedIdleLimit': time,
                'seedIdleMode': 1
            }

            post_data = {
                'arguments': arguments,
                'method': 'torrent-set'
            }

            if self._request(method='post', json=post_data):
                return self.response.json()['result'] == "success"
        else:
            return True

    def _set_torrent_priority(self, result):
        arguments = {'ids': [result.hash]}

        if result.priority == -1:
            arguments['priority-low'] = []
        elif result.priority == 1:
            # set high priority for all files in torrent
            arguments['priority-high'] = []
            # move torrent to the top if the queue
            arguments['queuePosition'] = 0
            if sickrage.app.config.torrent.high_bandwidth:
                arguments['bandwidthPriority'] = 1
        else:
            arguments['priority-normal'] = []

        post_data = {
            'arguments': arguments,
            'method': 'torrent-set'
        }

        if self._request(method='post', json=post_data):
            return self.response.json()['result'] == "success"

    def remove_torrent(self, info_hash):
        arguments = {
            'ids': [info_hash],
            'delete-local-data': 1,
        }

        post_data = {
            'arguments': arguments,
            'method': 'torrent-remove',
        }

        if self._request(method='post', json=post_data):
            return self.response.json()['result'] == "success"



"""Initial migration

Revision ID: 7
Revises:
Create Date: 2017-12-29 14:39:27.854291

"""
import datetime

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '19'
down_revision = '18'


def upgrade():
    conn = op.get_bind()
    meta = sa.MetaData(bind=conn)
    tv_shows = sa.Table('tv_shows', meta, autoload=True)

    with op.get_context().begin_transaction():
        for row in conn.execute(tv_shows.select()):
            conn.execute(f'UPDATE tv_shows SET last_refresh = "" WHERE tv_shows.indexer_id = {row.indexer_id}')
            conn.execute(f'UPDATE tv_shows SET last_xem_refresh = "" WHERE tv_shows.indexer_id = {row.indexer_id}')
            conn.execute(f'UPDATE tv_shows SET last_scene_exceptions_refresh = "" WHERE tv_shows.indexer_id = {row.indexer_id}')
            conn.execute(f'UPDATE tv_shows SET last_update = "" WHERE tv_shows.indexer_id = {row.indexer_id}')
            conn.execute(f'UPDATE tv_shows SET last_backlog_search = "" WHERE tv_shows.indexer_id = {row.indexer_id}')
            conn.execute(f'UPDATE tv_shows SET last_proper_search = "" WHERE tv_shows.indexer_id = {row.indexer_id}')

    with op.batch_alter_table("tv_shows") as batch_op:
        batch_op.alter_column('last_refresh', type_=sa.DateTime(timezone=True))
        batch_op.alter_column('last_xem_refresh', type_=sa.DateTime(timezone=True))
        batch_op.alter_column('last_scene_exceptions_refresh', type_=sa.DateTime(timezone=True))
        batch_op.alter_column('last_update', type_=sa.DateTime(timezone=True))
        batch_op.alter_column('last_backlog_search', type_=sa.DateTime(timezone=True))
        batch_op.alter_column('last_proper_search', type_=sa.DateTime(timezone=True))

    with op.get_context().begin_transaction():
        for row in conn.execute(tv_shows.select()):
            conn.execute(f'UPDATE tv_shows SET last_refresh = {sa.func.current_timestamp()} WHERE tv_shows.indexer_id = {row.indexer_id}')
            conn.execute(f'UPDATE tv_shows SET last_xem_refresh = {sa.func.current_timestamp()} WHERE tv_shows.indexer_id = {row.indexer_id}')
            conn.execute(f'UPDATE tv_shows SET last_scene_exceptions_refresh = {sa.func.current_timestamp()} WHERE tv_shows.indexer_id = {row.indexer_id}')
            conn.execute(f'UPDATE tv_shows SET last_update = {sa.func.current_timestamp()} WHERE tv_shows.indexer_id = {row.indexer_id}')
            conn.execute(f'UPDATE tv_shows SET last_backlog_search = {sa.func.current_timestamp()} WHERE tv_shows.indexer_id = {row.indexer_id}')
            conn.execute(f'UPDATE tv_shows SET last_proper_search = {sa.func.current_timestamp()} WHERE tv_shows.indexer_id = {row.indexer_id}')


def downgrade():
    pass



"""Initial migration

Revision ID: 2
Revises:
Create Date: 2017-12-29 14:39:27.854291

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '2'
down_revision = '1'


def upgrade():
    conn = op.get_bind()
    meta = sa.MetaData(bind=conn)
    tv_shows = sa.Table('tv_shows', meta, autoload=True)

    if not hasattr(tv_shows.c, 'last_backlog_search'):
        op.add_column('tv_shows', sa.Column('last_backlog_search', sa.Integer, default=0))


def downgrade():
    pass



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################

import re

from subliminal.video import Episode

import sickrage
from sickrage.core.common import Quality

SHOW_MAPPING = {
    'series_tvdb_id': 'tvdb_id',
    'series_imdb_id': 'imdb_id',
    'year': 'startyear'
}

EPISODE_MAPPING = {
    'tvdb_id': 'tvdb_id',
    'size': 'file_size',
    'title': 'name',
}

ADDITIONAL_MAPPING = {
    'season': 'season',
    'episode': 'episode',
    'release_group': 'release_group',
}

series_re = re.compile(r'^(?P<series>.*?)(?: \((?:(?P<year>\d{4})|(?P<country>[A-Z]{2}))\))?$')


def refine(video, tv_episode=None, **kwargs):
    """Refine a video by using TVEpisode information.

    :param video: the video to refine.
    :type video: Episode
    :param tv_episode: the TVEpisode to be used.
    :type tv_episode: medusa.tv.Episode
    :param kwargs:
    """
    if video.series_tvdb_id and video.tvdb_id:
        sickrage.app.log.debug('No need to refine with Episode')
        return

    if not tv_episode:
        sickrage.app.log.debug('No Episode to be used to refine')
        return

    if not isinstance(video, Episode):
        sickrage.app.log.debug(f'Video {video.name!r} is not an episode. Skipping refiner...')
        return

    if tv_episode.show:
        sickrage.app.log.debug('Refining using Series information.')
        series, year, _ = series_re.match(tv_episode.show.name).groups()
        enrich({'series': series, 'year': int(year) if year else None}, video)
        enrich(SHOW_MAPPING, video, tv_episode.show)

    sickrage.app.log.debug('Refining using Episode information.')
    enrich(EPISODE_MAPPING, video, tv_episode)
    enrich(ADDITIONAL_MAPPING, video, tv_episode, overwrite=False)
    guess = Quality.to_guessit(tv_episode.quality)
    enrich({'resolution': guess.get('screen_size'), 'source': guess.get('source')}, video, overwrite=False)


def enrich(attributes, target, source=None, overwrite=True):
    """Copy attributes from source to target.

    :param attributes: the attributes mapping
    :type attributes: dict(str -> str)
    :param target: the target object
    :param source: the source object. If None, the value in attributes dict will be used as new_value
    :param overwrite: if source field should be overwritten if not already set
    :type overwrite: bool
    """
    for key, value in attributes.items():
        old_value = getattr(target, key)
        if old_value and old_value != '' and not overwrite:
            continue

        new_value = getattr(source, value) if source else value

        if new_value and old_value != new_value:
            setattr(target, key, new_value)
            sickrage.app.log.debug(f'Attribute {key} changed from {old_value!r} to {new_value!r}')


# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################
import ipaddress
import json

import sentry_sdk
from tornado.web import authenticated

import sickrage
from sickrage.core.enums import UserPermission
from sickrage.core.helpers import get_internal_ip, get_external_ip, get_ip_address
from sickrage.core.webserver.handlers.base import BaseHandler


class AccountLinkHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        code = self.get_argument('code', None)

        redirect_uri = "{}://{}{}/account/link".format(self.request.protocol, self.request.host, sickrage.app.config.general.web_root)

        if code:
            token = sickrage.app.auth_server.authorization_code(code, redirect_uri)
            if not token:
                return self.redirect('/account/link')

            certs = sickrage.app.auth_server.certs()
            if not certs:
                return self.redirect('/account/link')

            decoded_token = sickrage.app.auth_server.decode_token(token['access_token'], certs)
            if not decoded_token:
                return self.redirect('/account/link')

            # if sickrage.app.api.token:
            #     sickrage.app.api.logout()

            exchanged_token = sickrage.app.auth_server.token_exchange(token['access_token'])
            if exchanged_token:
                sickrage.app.api.token = exchanged_token

            sickrage.app.config.general.enable_sickrage_api = True

            if not sickrage.app.config.user.sub_id:
                sickrage.app.config.user.sub_id = decoded_token.get('sub_id')
                sickrage.app.config.user.username = decoded_token.get('preferred_username')
                sickrage.app.config.user.email = decoded_token.get('email')
                sickrage.app.config.user.permissions = UserPermission.SUPERUSER

                sentry_sdk.set_user({
                    'id': sickrage.app.config.user.sub_id,
                    'username': sickrage.app.config.user.username,
                    'email': sickrage.app.config.user.email
                })

            if not sickrage.app.config.general.server_id:
                server_id = sickrage.app.api.server.register_server(
                    ip_addresses=','.join([get_internal_ip()]),
                    web_protocol=self.request.protocol,
                    web_port=sickrage.app.config.general.web_port,
                    web_root=sickrage.app.config.general.web_root,
                    server_version=sickrage.version()
                )

                if server_id:
                    sickrage.app.config.general.server_id = server_id

            if sickrage.app.config.general.server_id:
                sentry_sdk.set_tag('server_id', sickrage.app.config.general.server_id)

            sickrage.app.config.save(mark_dirty=True)

            sickrage.app.alerts.message(_('Linked SiCKRAGE account to SiCKRAGE API'))
        else:
            authorization_url = sickrage.app.auth_server.authorization_url(redirect_uri=redirect_uri, scope="profile email")
            if authorization_url:
                return self.redirect(authorization_url, add_web_root=False)

        return self.redirect('/account/link')


class AccountUnlinkHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        if not sickrage.app.config.general.server_id or sickrage.app.api.server.unregister_server(sickrage.app.config.general.server_id):
            if not sickrage.app.config.general.sso_auth_enabled:
                sickrage.app.config.reset_encryption()
                sickrage.app.config.general.server_id = ""
                sickrage.app.config.user.sub_id = ""

            del sickrage.app.api.token

            sickrage.app.config.general.enable_sickrage_api = False
            sickrage.app.config.save()

            sickrage.app.alerts.message(_('Unlinked SiCKRAGE account from SiCKRAGE API'))


class AccountIsLinkedHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        return json.dumps({'linked': ('true', 'false')[not sickrage.app.api.userinfo]})



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################
import os
import time
import traceback
from enum import Enum

import sickrage
from sickrage.core.process_tv import ProcessResult
from sickrage.core.queues import Queue, Task, TaskPriority, TaskStatus


class PostProcessorTaskActions(Enum):
    AUTO = 'Auto'
    MANUAL = 'Manual'


class PostProcessorQueue(Queue):
    def __init__(self):
        Queue.__init__(self, "POSTPROCESSORQUEUE")
        self._output = []

    @property
    def output(self):
        return '\n'.join(self._output)

    def log(self, message, level=None):
        sickrage.app.log.log(level or sickrage.app.log.INFO, message)
        self._output.append(message)

    def clear_log(self):
        self._output = []

    def find_in_queue(self, dirName, proc_type):
        """
        Finds any item in the queue with the given dirName and proc_type pair
        :param dirName: directory to be processed by the task
        :param proc_type: processing type, auto/manual
        :return: instance of PostProcessorTask or None
        """
        for task in self.tasks.copy().values():
            if isinstance(task, PostProcessorTask) and task.dirName == dirName and task.proc_type == proc_type:
                return True

        return False

    @property
    def queue_length(self):
        """
        Returns a dict showing how many auto and manual tasks are in the queue
        :return: dict
        """
        length = {'auto': 0, 'manual': 0}

        for task in self.tasks.copy().values():
            if isinstance(task, PostProcessorTask):
                if task.proc_type == 'auto':
                    length['auto'] += 1
                else:
                    length['manual'] += 1

        return length

    def put(self, dirName, nzbName=None, process_method=None, force=False, is_priority=None, delete_on=False,
            failed=False, proc_type="auto", force_next=False, **kwargs):
        """
        Adds an item to post-processing queue
        :param dirName: directory to process
        :param nzbName: release/nzb name if available
        :param process_method: processing method, copy/move/symlink/link
        :param force: force overwriting of existing files regardless of quality
        :param is_priority: whether to replace the file even if it exists at higher quality
        :param delete_on: delete files and folders after they are processed (always happens with move and auto combination)
        :param failed: mark downloads as failed if they fail to process
        :param proc_type: processing type: auto/manual
        :param force_next: wait until the current item in the queue is finished then process this item next
        :return: string indicating success or failure
        """

        self.clear_log()

        if not dirName:
            self.log("{} post-processing attempted but directory is not set".format(proc_type.title()), sickrage.app.log.WARNING)
            return self.output

        if not os.path.isabs(dirName):
            self.log("{} post-processing attempted but directory is relative (and probably not "
                     "what you really want to process): {}".format(proc_type.title(), dirName),
                     sickrage.app.log.WARNING)
            return self.output

        if not delete_on:
            delete_on = (False, (not sickrage.app.config.general.no_delete, True)[process_method == "move"])[proc_type == "auto"]

        if self.find_in_queue(dirName, proc_type):
            self.log("An item with directory {} is already being processed in the queue".format(dirName))
            return self.output
        else:
            task_id = super(PostProcessorQueue, self).put(
                PostProcessorTask(dirName, nzbName, process_method, force, is_priority, delete_on, failed, proc_type))

            if force_next:
                while self.check_status(task_id) not in [TaskStatus.FINISHED, TaskStatus.FAILED]:
                    time.sleep(1)

                return self.get_result(task_id)

            self.log("{} post-processing job for {} has been added to the queue".format(proc_type.title(), dirName))
            return self.output + "<p><span class='hidden'>Processing succeeded</span></p>"


class PostProcessorTask(Task):
    def __init__(self, dirName, nzbName=None, process_method=None, force=False, is_priority=None, delete_on=False, failed=False, proc_type="auto"):
        action = (PostProcessorTaskActions.MANUAL, PostProcessorTaskActions.AUTO)[proc_type == "auto"]
        super(PostProcessorTask, self).__init__(action.value, action)

        self.dirName = dirName
        self.nzbName = nzbName
        self.process_method = process_method
        self.force = force
        self.is_priority = is_priority
        self.delete_on = delete_on
        self.failed = failed
        self.proc_type = proc_type

        self.priority = (TaskPriority.HIGH, TaskPriority.NORMAL)[proc_type == 'auto']

    def run(self):
        """
        Runs the task
        :return: None
        """

        try:
            sickrage.app.log.info("Started {} post-processing job for: {}".format(self.proc_type, self.dirName))

            result = ProcessResult(self.dirName, self.process_method, self.proc_type).process(
                nzbName=self.nzbName,
                force=self.force,
                is_priority=self.is_priority,
                delete_on=self.delete_on,
                failed=self.failed
            )

            sickrage.app.log.info("Finished {} post-processing job for: {}".format(self.proc_type, self.dirName))
        except Exception:
            sickrage.app.log.debug(traceback.format_exc())
            result = '{}'.format(traceback.format_exc())
            result += 'Processing Failed'

        return result
